'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var alkali = require('alkali');
var alkali__default = _interopDefault(alkali);
var lmdb = require('lmdb');
var weakLruCache = require('weak-lru-cache');
var fs = require('fs');
var fs__default = _interopDefault(fs);
var crypto = require('crypto');
var net = require('net');
var path = require('path');
var msgpackr = require('msgpackr');
var stream = require('stream');

function when(promise, callback, errback) {
    if (promise && promise.then) {
        return errback ?
            promise.then(callback, errback) :
            promise.then(callback);
    }
    return callback(promise);
}

// this provides a weak-valued map to ensure we only have a single instance of an object per id, but can still be GC-ed
try {
    let allInstances = [];
    exports.WeakValueMap = function () {
        let map = new weakLruCache.WeakLRUCache({
            expirer: false,
        });
        allInstances.push(map);
        return map;
    };
    exports.WeakValueMap.getStatus = function () {
        let mapStats = [];
        for (let map of allInstances) {
            let size = 0;
            let count = 0;
            for (let key of map.keys()) {
                let value = map.get(key);
                size += value && value.approximateSize || 100;
                count++;
            }
            if (count > 0) {
                mapStats.push({
                    name: map.name,
                    size,
                    count
                });
            }
        }
        return mapStats;
    };
}
catch (error) {
    console.warn('No weak value map available, this can be used for development, but weak value maps should be enabled for production use', error.toString());
    exports.WeakValueMap = Map;
    exports.WeakValueMap.getStatus = function () {
        return 'WeakValueMap failed to load';
    };
    exports.WeakValueMap.prototype._keysAsArray = function () {
        return Array.from(this.keys());
    };
}

// a prioritized least recently used cache replacement/expiration strategy
// this is implemented as a probablistic multi-step descent towards expiration, with prioritized
// entry positions into the queue (allows repeatedly accessed, smaller objects
// to expire slower)
// number of entries in the cache. reduce this to reduce cache size (less memory)
let CACHE_ENTRIES = 30000;
// this is the mid-point size in the cache (roughly half of cache entries will
// be smaller, half larger). Reduce this to force large entries out faster (less memory)
const NOMINAL_SIZE = 100;
// this is the speed of decay towards expiration. Reducing this will result in
// more "accurate" expiration timing, but will also increase the overhead of the
// algorithm
const DECAY_RATE = 2.7;
const DECAY_INTERVAL = 10000;
const DECAY_REMOVAL = 10;
const EMPTY_SLOT = {
    set priority(priority) {
    },
    get priority() {
        return -1;
    },
};
class ExpirationStrategy {
    constructor() {
        this.cache = [];
        // periodically clean out entries so they decay over time as well.
        setInterval(() => {
            for (let i = 0; i < DECAY_REMOVAL; i++) {
                this.useEntry(EMPTY_SLOT, Math.random());
            }
        }, DECAY_INTERVAL).unref();
    }
    useEntry(entity, size) {
        if (!isFinite(size)) {
            size = 100;
        }
        let lastPriority = entity.priority;
        if (lastPriority > -1) {
            // remove from old slot if it is currently in cache
            this.cache[entity.priority] = null;
        }
        // define new priority
        // to prevent duplicate sizes from colliding, we add a revolving offset
        // this is an 8-bit revolving offset. We could add more bits if we wanted
        // to further reduce collisions, but minimizing offset bits actually helps leave "room"
        // for multi-occurence entries to stick around longer, and more offset bits could cause overlaps
        // in sizes which slightly reduces accuracies
        // offset = (offset + 157) & 255
        // calculate new priority/slot, placing large entries closer to expiration, smaller objects further from expiration
        let adjustedSize = size / NOMINAL_SIZE;
        let priority = entity.priority = Math.floor(CACHE_ENTRIES / (1 +
            (lastPriority > -1 ?
                (adjustedSize + CACHE_ENTRIES / lastPriority) / 3 :
                adjustedSize)));
        while (entity) {
            // iteratively place entries in cache, pushing old entries closer to the end of the queue
            priority = entity.priority = Math.floor(priority / DECAY_RATE);
            if (priority == -1) {
                return;
            }
            let entityToMove = this.cache[priority];
            this.cache[priority] = entity;
            entity = entityToMove;
            if (priority == 0 && entity) {
                // moved out of the queue, clear it from the cache
                entity.priority = -1; // keep priority a number type for more efficient class structure
                return;
            }
        }
    }
    deleteEntry(entity) {
        if (entity.priority > -1) {
            // remove from old slot if it is currently in cache
            this.cache[entity.priority] = null;
        }
    }
    isCached(entity) {
        return entity.priority > -1;
    }
    getSnapshot() {
        let totalSize = 0;
        let size;
        return {
            entries: this.cache.map(entity => entity && ({
                id: entity.id,
                type: entity.constructor.name,
                size: (size = entity.approximateSize, (totalSize += (size || 1000)), size)
            })),
            totalSize
        };
    }
    set cachedEntrySize(size) {
        CACHE_ENTRIES = size;
    }
    get cachedEntrySize() {
        return CACHE_ENTRIES;
    }
    clearEntireCache() {
        // clear the entire cache. This is useful for finding memory leaks
        for (let i in this.cache) {
            let entry = this.cache[i];
            if (entry) {
                this.cache[i] = null;
            }
        }
    }
}
ExpirationStrategy.defaultInstance = new ExpirationStrategy();
/*
This is the test I used to determine the optimal multiplier for spatial diversity:
for (var multiplier = 0; multiplier < 255; multiplier++) {
    entries = []
    sum = 0
    for (var i = 0; i < 256; i++) {
        let n = ((i * multiplier) & 255)
        for (var j = 0; j < 128; j++) {
            if (entries[(n - j + 256) & 255] || entries[(n + j + 256) & 255]) {
                sum += j
                //console.log(j)
                break
            }
        }
        entries[n] = true
    }
    if (sum > 800) {
        console.log(multiplier, sum)
    }
}
*/

class ExtendableError extends Error {
    constructor(message) {
        super(message);
        this.stack = (new Error()).stack;
        this.name = this.constructor.name;
    }
}
class AccessError extends ExtendableError {
    get status() {
        return 403;
    }
    get isExternal() {
        return true;
    }
}
class UnauthenticatedError extends ExtendableError {
    get status() {
        return 401;
    }
    get isExternal() {
        return true;
    }
}

const { spawn, UpdateEvent, currentContext } = alkali.Variable ? alkali : alkali__default;
let pipeServerStarted;
const classMap = new Map();
const streamByPidClass = new Map();
const streamsByClass = new Map();
const waitingRequests = new Map();
const whenProcessConnected = new Map();
const getPipePath = (processId) => path.join(path.sep == '/' ? '/tmp' : '\\\\?\\pipe', 'cobase-' + processId);
let nextRequestId = 1;
function startPipeClient(processId, Class) {
    let whenConnected;
    if (whenProcessConnected.has(processId)) {
        whenConnected = whenProcessConnected.get(processId);
    }
    else {
        whenConnected = new Promise((resolve, reject) => {
            const tryToConnect = (retries) => {
                const socket = net.createConnection(getPipePath(processId));
                let parsedStream = socket.pipe(new msgpackr.UnpackrStream()).on('error', (error) => {
                    console.error('Error in pipe client socket', error);
                });
                let serializingStream = new msgpackr.PackrStream();
                serializingStream.pipe(socket);
                serializingStream.pid = processId;
                let connected;
                socket.on('error', (error) => {
                    if (connected)
                        console.error(error); // shouldn't happen after a connection
                    else {
                        if (retries > 2)
                            reject(error);
                        else
                            setTimeout(() => {
                                tryToConnect(retries + 1);
                            }, 1500);
                    }
                }).on('connect', () => {
                    connected = true;
                    console.debug('Connected to process', processId);
                    resolve(serializingStream);
                });
                socket.on('close', (event) => {
                    serializingStream.emit('close', event);
                });
                socket.unref();
                parsedStream.on('data', (message) => {
                    onMessage(message, serializingStream);
                });
            };
            tryToConnect(0);
        });
        whenProcessConnected.set(processId, whenConnected);
    }
    return whenConnected.then(stream => {
        attachClass(stream, Class, processId);
        // declare this class listens on this stream, by sending out a process identification
        stream.write({
            className: Class.name,
            pid: process.pid
        });
    });
}
function startPipeServer() {
    if (pipeServerStarted)
        return;
    pipeServerStarted = true;
    net.createServer((socket) => {
        socket.pipe(new msgpackr.UnpackrStream()).on('data', (message) => {
            onMessage(message, serializingStream);
        });
        let serializingStream = new msgpackr.PackrStream();
        socket.on('close', (event) => {
            serializingStream.emit('close', event);
        });
        serializingStream.pipe(socket);
        serializingStream.isIncoming = true;
    }).on('error', (err) => {
        // handle errors here
        throw err;
    }).listen(getPipePath(process.pid));
}
function attachClass(stream, Class, processId) {
    stream.pid = processId;
    const className = Class.name;
    let streams = streamsByClass.get(className);
    if (!streams) {
        streamsByClass.set(className, streams = []);
    }
    streams.push(stream);
    streamByPidClass.set(processId + '-' + className, stream);
    const otherProcesses = Class.otherProcesses || (Class.otherProcesses = []);
    if (!otherProcesses.includes(processId) && processId !== process.pid) {
        otherProcesses.push(processId);
    }
    let queue = [];
    let timeoutToWrite;
    const updater = {
        updated(event, by) {
            // TODO: debounce
            //console.log('sending update event', className, process.pid)
            let id = by && by.id;
            if (id && event.source && by.constructor === event.source.constructor && !event.sourceProcess) {
                try {
                    const eventToSerialize = Object.assign({}, event, {
                        instanceId: id,
                        method: 'updated',
                        className,
                        type: event.type,
                    });
                    delete eventToSerialize.visited;
                    delete eventToSerialize.source;
                    if (eventToSerialize.sources) {
                        eventToSerialize.sources = Array.from(eventToSerialize.sources).map(source => ({
                            id: source.id,
                            typeName: source.constructor.name,
                        }));
                    }
                    delete eventToSerialize.previousValues;
                    delete eventToSerialize.target;
                    delete eventToSerialize.oldValue;
                    delete eventToSerialize.whenWritten;
                    when(event.whenWritten, () => {
                        queue.push(eventToSerialize);
                        if (!timeoutToWrite)
                            timeoutToWrite = setTimeout(() => {
                                timeoutToWrite = null;
                                stream.write(queue);
                                queue = [];
                            }, 10);
                    });
                }
                catch (error) {
                    // TODO: Not sure how we get in this state
                    console.warn(error);
                    Class.stopNotifies(updater);
                }
            }
        },
        stream,
        Class
    };
    Class.notifies(updater);
    Class.sendBroadcast = notification => {
        for (const stream of streams) {
            notification.className = className;
            stream.write(notification);
        }
    };
    Class.sendRequestToProcess = (pid, message) => {
        const requestId = message.requestId = nextRequestId++;
        message.className = Class.name;
        const stream = streamByPidClass.get(pid + '-' + className);
        if (!stream) {
            return; // TODO: If it is undefined wait for a connection
        }
        stream.write(message);
        return new Promise((resolve, reject) => waitingRequests.set(requestId, { resolve, reject }));
    };
    Class.sendRequestToAllProcesses = (message) => {
        message.className = Class.name;
        return Promise.all(streams.map(stream => {
            const requestId = message.requestId = nextRequestId++;
            stream.write(message);
            return new Promise((resolve, reject) => waitingRequests.set(requestId, { resolve, reject }));
        }));
    };
    stream.setMaxListeners(1000); // we are going to be adding a lot here
    stream.on('close', () => {
        Class.stopNotifies(updater);
        streams.splice(streams.indexOf(stream), 1);
        let otherProcessIndex = otherProcesses.indexOf(processId);
        if (otherProcessIndex > -1)
            otherProcesses.splice(otherProcessIndex, 1);
        streamByPidClass.delete(processId + '-' + className);
    });
}
function onMessage(message, stream) {
    try {
        if (message instanceof Array) {
            for (let part of message)
                onMessage(part, stream);
            return;
        }
        const { requestId, responseId, className, instanceId } = message;
        if (responseId) {
            const resolver = waitingRequests.get(responseId);
            waitingRequests.delete(responseId);
            return resolver.resolve(message);
        }
        let target = classMap.get(className);
        if (target) {
            if (requestId) {
                try {
                    when(target.receiveRequest(message), (result) => {
                        result = result || {};
                        result.responseId = requestId;
                        stream.write(result);
                    }, (error) => {
                        stream.write({
                            error,
                            responseId: requestId
                        });
                    });
                }
                catch (error) {
                    stream.write({
                        error,
                        responseId: requestId
                    });
                }
            }
            else {
                if (message.type) {
                    const event = new UpdateEvent();
                    event.sourceProcess = stream.pid;
                    event.source = target;
                    Object.assign(event, message);
                    if (message.sources) {
                        event.sources = message.sources.map(source => ({
                            id: source.id,
                            constructor: classMap.get(source.typeName)
                        }));
                    }
                    target.updated(event, instanceId && { id: instanceId });
                }
                else if (message.pid) {
                    attachClass(stream, target, message.pid);
                }
                else {
                    target.update(message);
                }
            }
        }
        else {
            console.warn('Unknown message received', message);
        }
    }
    catch (error) {
        console.error('Handling message error', error);
    }
}
function registerClass(Class) {
    startPipeServer(); // Maybe start it in the next event turn so you can turn it off in single process environment?
    classMap.set(Class.name, Class);
}
function addProcess(pid, Class) {
    return startPipeClient(pid, Class);
}
/*function onCloseSocket(stream, processId) {
    const pid = stream.pid
    let index = streams.indexOf(stream)
    if (index > -1)
        streams.splice(index, 1)
    let removed = 0
    for (let updater of updaters) {
        if (updater.stream === stream) {
            //console.log('stop notifications for', process.pid, 'from', pid)
            removed++
            updater.Class.stopNotifies(updater)
            index = updaters.indexOf(updater)
            if (index > -1)
                updaters.splice(index, 1)
        }
    }
    console.log('socket close from', process.pid, 'to', processId, pid, 'removed updaters', removed)
    streamByPid.set(pid, null) // set it to null, so we now it once existed and is dead
}

/*
// every child process should be ready to join the network
process.on('message', (data) => {
    if (data.enterNetwork) {
        console.log('Received request to start pipe server')
        // create pipe server
        startPipeServer()
        // need to send confirmation that it is set up.
        process.send({
            enteredNetwork: true
        })
    } else if (data.connectToProcess) {
        startPipeClient(data.connectToProcess)
    }
})
*/

const { Context } = alkali.Context ? alkali : alkali__default;
class RequestContext extends Context {
    constructor(request, session, user) {
        super(user);
        this.request = request;
        this.session = session;
    }
    newContext() {
        return new this.constructor(this.request, this.session, this.subject);
    }
    get expectedVersions() {
        const session = this.session || (this.session = {});
        return session.expectedVersions || (session.expectedVersions = {});
    }
}
function setRequestContextClass(Context) {
}
const DEFAULT_CONTEXT = {
    expectedVersions: {}
};

const { Transform, VPromise, VArray, Variable, spawn: spawn$1, currentContext: currentContext$1, NOT_MODIFIED, getNextVersion, ReplacedEvent, DeletedEvent, AddedEvent, UpdateEvent: UpdateEvent$1, Context: Context$1 } = alkali.Variable ? alkali : alkali__default;
let getCurrentContext = () => currentContext$1;
const DEFAULT_INDEXING_DELAY = 20;
const DEFAULT_INDEXING_CONCURRENCY = 20;
const expirationStrategy = ExpirationStrategy.defaultInstance;
const instanceIdsMap = new exports.WeakValueMap();
const DB_VERSION_KEY = Buffer.from([1, 1]); // table metadata
const INITIALIZING_PROCESS_KEY = Buffer.from([1, 4]);
// everything after 9 is cleared when a db is cleared
const SHARED_STRUCTURE_KEY = Buffer.from([1, 10]);
const INITIALIZING_LAST_KEY = Buffer.from([1, 7]);
const DISCOVERED_SOURCE = 'is-discovered';
const INDEXING_MODE = {};
const NO_CACHE = 2;
const AS_SOURCE = {};
const EXTENSION = '.mdb';
const DB_FORMAT_VERSION = 0;
const allStores = new Map();
let globalDoesInitialization;
global.cache = expirationStrategy; // help with debugging
class InstanceIds extends Transform.as(VArray) {
    transform() {
        return when(when(this.Class.resetProcess, () => this.Class.whenWritten), () => this.Class.getInstanceIds().asArray);
    }
    getValue() {
        return when(super.getValue(true), ids => {
            expirationStrategy.useEntry(this, ids.length);
            return ids;
        });
    }
    valueOf() {
        return super.valueOf(true); // always allow promises to be returned
    }
    clearCache() {
        this.cachedValue = undefined;
        this.cachedVersion = -1;
    }
}
const MakePersisted = (Base) => {
    var _a;
    return secureAccess((_a = class extends Base {
            constructor(id) {
                super();
                if (id == null) {
                    throw new TypeError('No id provided');
                }
                if (this.constructor === Persisted) {
                    throw new TypeError('Can not directly instantiate abstract Persisted class');
                }
                if (this.checkSourceVersions)
                    this.readyState = 'invalidated'; // start in this state for items that might not be updated so freshly loaded entities don't bypass version checks
                this.id = id;
                let sources = this.constructor.sources;
                if (sources) {
                    for (let i = 0; i < sources.length; i++) {
                        this['source' + (i ? i : '')] = sources[i].for(id);
                    }
                }
            }
            get staysUpdated() {
                return true;
            }
            static get defaultInstance() {
                return this._defaultInstance || (this._defaultInstance = new Variable());
            }
            static for(id) {
                if (id > 0 && typeof id === 'string' || id == null) {
                    throw new Error('Id should be a number or non-numeric string: ' + id + 'for ' + this.name);
                }
                let instancesById = this.instancesById;
                if (!instancesById) {
                    this.ready;
                    instancesById = this.instancesById;
                }
                let instance = instancesById.getValue(id);
                if (!instance) {
                    instance = new this(id);
                    instancesById.setValue(id, instance);
                }
                return instance;
            }
            static getByIds(ids) {
                // for optimized access to a set of ids
                if (!(ids instanceof Array))
                    ids = Array.from(ids);
                let i = 0, l = ids.length;
                let values = [];
                let promised = [];
                const getNext = () => {
                    while (i < l) {
                        let value = this.for(ids[i]);
                        if (value && value.then) {
                            // a promise, put in the list of parallel promises
                            let promisedI = i++;
                            promised.push(value.then(value => {
                                values[promisedI] = value;
                            }));
                            if (promised.length > (this.MAX_CONCURRENCY || 100)) {
                                let promisedToFinish = promised;
                                promised = [];
                                return Promise.all(promisedToFinish).then(getNext);
                            }
                        }
                        else {
                            values[i++] = value;
                        }
                    }
                    if (promised.length > 0) {
                        return Promise.all(promised);
                    }
                };
                return when(getNext(), () => values);
            }
            static index(propertyName, indexBy) {
                let index = this['index-' + propertyName];
                if (index) {
                    return index;
                }
                index = this['index-' + propertyName] = class extends Index({ Source: this }) {
                    static indexBy(entity, sourceKey) {
                        return indexBy ? indexBy(entity, sourceKey) : entity[propertyName];
                    }
                };
                Object.defineProperty(index, 'name', { value: this.name + '-index-' + propertyName });
                index.start();
                return index;
            }
            /*	static with(properties) {
                    let DerivedClass = super.with(properties)
                    DerivedClass.sources = [this]
                    let hasRelatedProperties
                    for (let key of properties) {
                        let property = properties[key]
                        if (property.initialized) {
                            property.initialized(this)
                            hasRelatedProperties = true
                        }
            
                    }
                    if (hasRelatedProperties) {
                        DerivedClass.prototype.transform = function(data, ...propertySources) {
                            for (let propertySource of propertySources) {
                                data[DerivedClass.sources[i].key] = propertySource
                            }
                            return data
                        }
                    }
                    return DerivedClass
                }*/
            static relatesBy(foreignKey) {
                let TargetClass = this;
                function relatesBy() { }
                relatesBy.defineAs = function (propertyName, Parent) {
                    let RelatedIndex = TargetClass.index(foreignKey);
                    let sourceIndex = Parent.sources.push(RelatedIndex) - 1;
                    let existingTransform = Parent.prototype.transform;
                    Parent.prototype.transform = function (primaryData) {
                        if (existingTransform) {
                            primaryData = existingTransform.apply(this, arguments);
                        }
                        let source = arguments[sourceIndex];
                        return Object.assign({ [propertyName]: source }, primaryData);
                    };
                    Parent.assign({
                        [propertyName]: VArray.of(TargetClass)
                    });
                };
                return relatesBy;
            }
            static relatedBy(foreignKey) {
                let TargetClass = this;
                function relatedBy() { }
                relatedBy.defineAs = function (propertyName, Parent) {
                    let ParentSource = Parent.sources[0];
                    let RelatedIndex = ParentSource.index(foreignKey);
                    let existingTransform = Parent.prototype.transform;
                    Parent.prototype.transform = function (primaryData) {
                        if (existingTransform) {
                            primaryData = existingTransform.apply(this, arguments);
                        }
                        return when(primaryData, primaryData => {
                            let reference = foreignKey.call ? foreignKey(primaryData) : primaryData[foreignKey];
                            return (reference instanceof Array ?
                                Promise.all(reference.map(ref => TargetClass.for(ref))) :
                                TargetClass.for(reference)).then(relatedValue => Object.assign({ [propertyName]: relatedValue }, primaryData));
                        });
                    };
                    TargetClass.notifies({
                        updated(event, by) {
                            RelatedIndex.for(by.id).getIndexedKeys().map(fromId => {
                                Parent.for(fromId).updated(event);
                            }).resolveData();
                        }
                    });
                    Parent.assign({
                        [propertyName]: TargetClass
                    });
                };
                return relatedBy;
            }
            static cacheWith(properties) {
                const CachedWith = Cached.from(this).assign(properties);
                Object.defineProperty(CachedWith, 'name', {
                    value: this.name + '-with-' + Object.keys(properties).filter(key => properties[key] && properties[key].defineAs).join('-')
                });
                CachedWith.start();
                return CachedWith;
            }
            transform(source) {
                return source;
            }
            static updatesRecorded(event) {
                return (event && event.updatesInProgress) ? Promise.all(event.updatesInProgress) : Promise.resolve();
            }
            delete() {
                return this.constructor.remove(this.id);
            }
            reset(action) {
                this.updated();
            }
            static get ready() {
                return this.start();
            }
            static start() {
                if (!this.hasOwnProperty('_ready')) {
                    this._ready = Promise.resolve(this.initialize());
                    this._ready.then(() => {
                        //console.log(this.name, 'is ready and initialized')
                        this.initialized = true;
                    }, (error) => {
                        console.error('Error initializing', this.name, error);
                    });
                }
                return this._ready;
            }
            static clearAllData() {
                let db = this.db;
                db.transactionSync(() => {
                    db.clear();
                });
                console.debug('Cleared the database', this.name, ', rebuilding');
            }
            static register(sourceCode) {
                // check the transform hash
                if (sourceCode) {
                    let moduleFilename = sourceCode.id || sourceCode;
                    if (sourceCode.version) {
                        // manually provide hash
                        this.version = sourceCode.version;
                    }
                    else if (typeof moduleFilename == 'string') {
                        // create a hash from the module source
                        this.version = fs.statSync(moduleFilename).mtime.getTime();
                        let hmac = crypto.createHmac('sha256', 'cobase');
                        hmac.update(fs.readFileSync(moduleFilename, { encoding: 'utf8' }));
                        this.transformHash = hmac.digest('hex');
                    }
                }
                return this.ready;
            }
            static get doesInitialization() {
                return this._doesInitialization === undefined ? globalDoesInitialization : this._doesInitialization;
            }
            static set doesInitialization(flag) {
                this._doesInitialization = flag;
            }
            static initializeRootDB() {
                const db = this.rootDB;
                this.rootStore = this;
                // TODO: Might be better use Buffer.allocUnsafeSlow(6)
                const processKey = this.processKey = Buffer.from([1, 3, (process.pid >> 24) & 0xff, (process.pid >> 16) & 0xff, (process.pid >> 8) & 0xff, process.pid & 0xff]);
                let initializingProcess;
                db.transactionSync(() => {
                    initializingProcess = db.get(INITIALIZING_PROCESS_KEY);
                    initializingProcess = initializingProcess && +initializingProcess.toString();
                    this.otherProcesses = Array.from(db.getRange({
                        start: Buffer.from([1, 3]),
                        end: INITIALIZING_PROCESS_KEY,
                    }).map(({ key, value }) => (key[2] << 24) + (key[3] << 16) + (key[4] << 8) + key[5])).filter(pid => !isNaN(pid));
                    db.putSync(processKey, '1'); // register process, in ready state
                    if ((!initializingProcess || !this.otherProcesses.includes(initializingProcess)) && this.doesInitialization !== false) {
                        initializingProcess = null;
                        db.putSync(INITIALIZING_PROCESS_KEY, process.pid.toString());
                    }
                    if (this.otherProcesses.includes(process.pid)) {
                        //console.warn('otherProcesses includes self')
                        this.otherProcesses.splice(this.otherProcesses.indexOf(process.pid));
                    }
                });
                this.initializingProcess = initializingProcess;
                this.whenUpgraded = Promise.resolve();
                const waitForUpgraded = () => {
                    let whenUpgraded = this.whenUpgraded;
                    whenUpgraded.then(() => setTimeout(() => {
                        if (whenUpgraded == this.whenUpgraded)
                            try {
                                this.db.removeSync(INITIALIZING_PROCESS_KEY);
                            }
                            catch (error) {
                                console.warn(error.toString());
                            }
                        else
                            return waitForUpgraded();
                    }), 10);
                };
                waitForUpgraded();
            }
            static getStructureVersion() {
                // default version handling is just to get the static version and hash with source versions, but this can be overriden with something
                // that gets this asynchronously or uses other logic
                let aggregateVersion = 0;
                for (let Source of this.sources || []) {
                    let version = Source.getStructureVersion && Source.getStructureVersion() || 0;
                    aggregateVersion = (aggregateVersion ^ version) * 1049011 + (aggregateVersion / 5555555 >>> 0);
                }
                return aggregateVersion ^ (this.version || 0);
            }
            static openDatabase() {
                const options = {
                    compression: true,
                    useFloat32: 3,
                    sharedStructuresKey: SHARED_STRUCTURE_KEY,
                    cache: true,
                    noMemInit: true,
                    useWritemap: false,
                };
                if (this.maxSharedStructures)
                    options.maxSharedStructures = this.maxSharedStructures;
                if (this.shouldShareStructure)
                    options.shouldShareStructure = this.shouldShareStructure;
                if (this.maxDbs)
                    options.maxDbs = this.maxDbs;
                if (this.useWritemap)
                    options.useWritemap = this.useWritemap;
                if (this.useFloat32)
                    options.useFloat32 = this.useFloat32;
                if (clearOnStart) {
                    console.info('Completely clearing', this.name);
                    options.clearOnStart = true;
                }
                this.rootDB = lmdb.open(this.dbFolder + '/' + this.name + EXTENSION, options);
                Object.assign(this, this.rootDB.get(DB_VERSION_KEY));
                this.prototype.db = this.db = this.openDB(this, { useVersions: true, cache: true });
                return true;
            }
            static initialize() {
                this.instancesById = new (this.useWeakMap ? exports.WeakValueMap : Map)();
                clearTimeout(this._registerTimeout);
                if (allStores.get(this.name)) {
                    throw new Error(this.name + ' already registered');
                }
                if (!storesObject[this.name])
                    storesObject[this.name] = this;
                allStores.set(this.name, this);
                for (let Source of this.sources || []) {
                    if (Source.start)
                        Source.start();
                    Source.notifies(this);
                }
                let isRoot = this.openDatabase();
                this.instancesById.name = this.name;
                return when(this.getStructureVersion(), structureVersion => {
                    this.expectedDBVersion = (structureVersion || 0) ^ (DB_FORMAT_VERSION << 12);
                    if (isRoot)
                        this.initializeRootDB();
                    let initializingProcess = this.rootStore.initializingProcess;
                    const db = this.db;
                    registerClass(this);
                    let whenEachProcess = [];
                    for (const pid of this.otherProcesses) {
                        whenEachProcess.push(addProcess(pid, this).catch(() => this.cleanupDeadProcessReference(pid, initializingProcess)));
                    }
                    // make sure these are inherited
                    if (initializingProcess || this.doesInitialization === false) {
                        // there is another process handling initialization
                        return when(whenEachProcess.length > 0 && Promise.all(whenEachProcess), (results) => {
                            console.debug('Connected to each process complete and finished reset initialization');
                        });
                    }
                    return this.doDataInitialization();
                }, (error) => {
                    console.error('Error getting database version', error);
                });
            }
            static doDataInitialization() {
                let whenThisUpgraded;
                try {
                    whenThisUpgraded = when(this.initializeData(), () => {
                    }, (error) => {
                        console.error('Error initializing database for', this.name, error);
                    });
                }
                catch (error) {
                    console.error(error);
                    whenThisUpgraded = Promise.resolve();
                }
                this.rootStore.whenUpgraded = this.rootStore.whenUpgraded.then(() => whenThisUpgraded);
                return whenThisUpgraded;
            }
            static cleanupDeadProcessReference(pid, initializingProcess) {
                // error connecting to another process, which means it is dead/old and we need to clean up
                // and possibly take over initialization
                let index = this.otherProcesses.indexOf(pid);
                const db = this.rootDB;
                if (index > -1) {
                    this.otherProcesses.splice(index, 1);
                    let deadProcessKey = Buffer.from([1, 3, (pid >> 24) & 0xff, (pid >> 16) & 0xff, (pid >> 8) & 0xff, pid & 0xff]);
                    let invalidationState = db.get(deadProcessKey);
                    if (this.doesInitialization !== false) {
                        db.transactionSync(async () => {
                            db.removeSync(deadProcessKey);
                        });
                    }
                }
                if (initializingProcess == pid && this.doesInitialization !== false) {
                    db.transactionSync(() => {
                        // make sure it is still the initializing process
                        initializingProcess = db.get(Buffer.from([1, 4]));
                        initializingProcess = initializingProcess && +initializingProcess.toString();
                        if (initializingProcess == pid) {
                            // take over the initialization process
                            //					console.log('Taking over initialization of', this.name, 'from process', initializingProcess)
                            initializingProcess = process.pid;
                            db.putSync(INITIALIZING_PROCESS_KEY, initializingProcess.toString());
                        }
                    });
                    if (initializingProcess == process.pid) {
                        return this.doDataInitialization();
                    }
                }
            }
            static async reset() {
                if (this.sources && this.sources[0])
                    this.sources[0].wasReset = false;
                this.clearAllData();
                await this.resetAll();
                this.updateDBVersion();
                this.resumeQueue();
            }
            static async initializeData() {
                const db = this.db;
                this.state = 'initializing data';
                //console.log('comparing db versions', this.name, this.dbVersion, this.expectedDBVersion)
                if (this.dbVersion == this.expectedDBVersion) ;
                else {
                    console.log('transform/database version mismatch, reseting db table', this.name, this.expectedDBVersion, this.dbVersion, this.version);
                    this.wasReset = true;
                    this.startVersion = getNextVersion();
                    const clearDb = !!this.dbVersion; // if there was previous state, clear out all entries
                    this.clearAllData();
                    await this.resetAll();
                    this.updateDBVersion();
                }
                let readyPromises = [];
                for (let Source of this.sources || []) {
                    readyPromises.push(Source.ready);
                }
                await Promise.all(readyPromises);
                this.resumePromise = this.resumeQueue(); // don't wait for this, it has its own separate promise system
            }
            valueOf(mode) {
                return super.valueOf(mode || true);
            }
            getValue(mode) {
                return this.constructor.get(this.id, mode);
            }
            gotValue(value) {
                // bypass any variable checks, since the data is coming from a DB
                return value;
            }
            updated(event = new ReplacedEvent(), by) {
                if (!event.visited) {
                    event.visited = new Set(); // TODO: Would like to remove this at some point
                }
                if (!event.source) {
                    event.source = this;
                }
                let context = getCurrentContext();
                if (context && !event.triggers && context.connectionId) {
                    event.triggers = [context.connectionId];
                }
                let Class = this.constructor;
                super.updated(event, by);
                Class.updated(event, this); // main handling occurs here
                // notify class listeners too
                return event;
            }
            static instanceSetUpdated(event) {
                let instanceIds = instanceIdsMap.getValue(this.name);
                if (instanceIds) {
                    instanceIds.updated(event);
                }
            }
            static clearEntryCache(id) {
                let entryCache = this._entryCache;
                if (entryCache) {
                    let entry = entryCache.get(id);
                    if (entry !== undefined) {
                        expirationStrategy.deleteEntry(entry);
                        entryCache.delete(id);
                    }
                }
            }
            static invalidateEntry(id, event) {
                this.clearEntryCache(id);
            }
            static update(id, event) {
                // this an easier way to manually call the updated process
                return this.updated(new ReplacedEvent(), { id });
            }
            static updated(event = new ReplacedEvent(), by) {
                if (!event.visited) {
                    event.visited = new Set(); // TODO: Would like to remove this at some point
                }
                if (event.visited.has(this)) {
                    return event;
                }
                event.visited.add(this);
                let context = getCurrentContext();
                if (context && !event.triggers && context.connectionId) {
                    event.triggers = [context.connectionId];
                }
                if (event && !event.version) {
                    event.version = getNextVersion();
                }
                let id = by && by.id;
                let nextBy = {
                    id,
                    constructor: this
                };
                if (!event.source) {
                    event.source = nextBy;
                }
                if (event.type === 'discovered' || event.type === 'added' || event.type === 'deleted') {
                    this.instanceSetUpdated(event);
                }
                if (event.type === 'reload-entry') {
                    if (event.doUpdate && by.constructor == this)
                        this.invalidateEntry(id, event, by);
                }
                else if (event.type === 'discovered') ;
                else if (id) {
                    this.invalidateEntry(id, event, by);
                }
                if (id) {
                    let instance;
                    instance = this.instancesById.getValue(id);
                    if (instance)
                        instance.updated(event, nextBy);
                }
                for (let listener of this.listeners || []) {
                    listener.updated(event, nextBy);
                }
                if (!context || !context.expectedVersions) {
                    context = DEFAULT_CONTEXT;
                }
                context.expectedVersions[this.name] = event.version;
                const whenUpdateProcessed = event.whenUpdateProcessed;
                if (whenUpdateProcessed) {
                    this.whenUpdateProcessed = whenUpdateProcessed;
                }
                return event;
            }
            static saveDBVersions() {
                this.rootDB.putSync(DB_VERSION_KEY, {
                    dbVersion: this.dbVersion,
                    childStores: this.childStores && this.childStores.map && this.childStores.map(childStore => ({
                        name: childStore.name,
                        dbVersion: childStore.dbVersion,
                    }))
                });
            }
            static updateDBVersion() {
                let version = this.startVersion;
                this.dbVersion = this.expectedDBVersion;
                console.debug('saving db version', this.name, this.dbVersion);
                this.rootStore.saveDBVersions();
                return version;
            }
            notifies(target) {
                let context = getCurrentContext();
                if (context) {
                    (this.listenersWithContext || (this.listenersWithContext = new Map())).set(target, context);
                }
                return super.notifies(target);
            }
            stopNotifies(target) {
                // standard variable handling
                if (this.listenersWithContext) {
                    this.listenersWithContext.delete(target);
                }
                return super.stopNotifies(target);
            }
            init() {
                if (!this.subscribedInstances) {
                    this.subscribedInstances = new Map();
                }
                this.subscribedInstances.set(this.id, this);
                return super.init();
            }
            cleanup() {
                this.subscribedInstances.delete(this.id);
                return super.cleanup();
            }
            static notifies(target) {
                let context = getCurrentContext();
                if (context) {
                    (this.listenersWithContext || (this.listenersWithContext = new Map())).set(target, context);
                }
                // standard variable handling (don't use alkali's contextual notifies)
                return Variable.prototype.notifies.call(this, target);
            }
            static stopNotifies(target) {
                // standard variable handling
                if (this.listenersWithContext) {
                    this.listenersWithContext.delete(target);
                }
                return Variable.prototype.stopNotifies.call(this, target);
            }
            static whenUpdatedInContext(waitForIndexing) {
                // transitively wait on all sources that need to update to this version
                let promises = [];
                for (let Source of this.sources || []) {
                    let whenUpdated = Source.whenUpdatedInContext && Source.whenUpdatedInContext();
                    if (whenUpdated && whenUpdated.then) {
                        promises.push(whenUpdated);
                    }
                }
                let whenReady;
                if (promises.length > 1) {
                    whenReady = Promise.all(promises);
                }
                else if (promises.length == 1) {
                    whenReady = promises[0];
                }
                if (waitForIndexing) {
                    let currentContext = getCurrentContext();
                    let updateContext = (currentContext && currentContext.expectedVersions) ? currentContext : DEFAULT_CONTEXT;
                    return when(whenReady, () => this.whenIndexedAndCommitted); /*{
                        if (updateContext.expectedVersions && updateContext.expectedVersions[this.name] > this.lastIndexedVersion && this.queue && this.queue.size > 0) {
                            // if the expected version is behind, wait for processing to finish
                            return this.requestProcessing(1) // up the priority
                        }
                    })*/
                }
                return whenReady;
            }
            static get instanceIds() {
                let instanceIds = instanceIdsMap.getValue(this.name);
                if (!instanceIds) {
                    instanceIdsMap.setValue(this.name, instanceIds = new InstanceIds());
                    instanceIds.Class = this;
                }
                return instanceIds;
            }
            exclusiveLock(executeWithLock) {
                let promisedResult;
                if (this.currentLock) {
                    let context = getCurrentContext();
                    const executeInContext = () => context.executeWithin(executeWithLock);
                    promisedResult = this.currentLock.then(executeInContext, executeInContext);
                }
                else {
                    let result = executeWithLock();
                    if (result && result.then)
                        promisedResult = result;
                    else
                        return result;
                }
                let thisLock, sync;
                const afterExecution = () => {
                    if (thisLock === this.currentLock) {
                        this.currentLock = null;
                    }
                    sync = true;
                };
                thisLock = this.currentLock = promisedResult.then(afterExecution, (error) => {
                    // Probably need to review if uncaught promise rejections are properly handled
                    console.error(error);
                    afterExecution();
                });
                if (sync) {
                    this.currentLock = null;
                }
                return promisedResult;
            }
            static tryForQueueEntry(id, action) {
                this.lastIndexingId = id;
                const onQueueError = async (error) => {
                    let indexRequest = this.queue && this.queue.get(id) || {};
                    let version = indexRequest.version;
                    if (error.isTemporary) {
                        let retries = indexRequest.retries = (indexRequest.retries || 0) + 1;
                        this.state = 'retrying index in ' + retries * 1000 + 'ms';
                        if (retries < (this.maxRetries || 1000)) {
                            this.isRetrying = true;
                            await delay(retries * 1000);
                            this.isRetrying = false;
                            //console.info('Retrying index entry', this.name, id, error)
                            return this.tryForQueueEntry(id, action);
                        }
                        else {
                            console.info('Too many retries', this.name, id, retries);
                        }
                    }
                    if (indexRequest && indexRequest.version !== version)
                        return; // if at any point it is invalidated, break out, don't log errors from invalidated states
                    console.warn('Error indexing', this.name, id, error);
                    if (this.queue && this.queue.delete)
                        this.queue.delete(id); // give up and delete it
                };
                try {
                    let result = action(id);
                    if (result && result.catch)
                        return result.catch(error => onQueueError(error));
                }
                catch (error) {
                    return onQueueError(error);
                }
            }
            static async processQueue(queue) {
                this.state = 'processing';
                if (this.onStateChange) {
                    this.onStateChange({ processing: true, started: true });
                }
                let cpuUsage = process.cpuUsage();
                let cpuTotalUsage = cpuUsage.user + cpuUsage.system;
                try {
                    queue = queue || this.queue;
                    let initialQueueSize = queue.size;
                    //currentlyProcessing.add(this)
                    if (initialQueueSize > 100 || initialQueueSize == undefined) {
                        console.log('Indexing', initialQueueSize || '', this.name, 'for', this.name);
                    }
                    let actionsInProgress = new Set();
                    this.actionsInProgress = actionsInProgress;
                    let sinceLastStateUpdate = 0;
                    let lastTime = Date.now();
                    let delayMs = 10;
                    let indexed = 0;
                    do {
                        if (this.nice > 0)
                            await delay(this.nice); // short delay for other processing to occur
                        for (let [id] of queue) {
                            if (queue.isReplaced)
                                return;
                            sinceLastStateUpdate++;
                            this.state = 'indexing entry ' + id;
                            let now = Date.now();
                            indexed++;
                            let desiredConcurrentRatio = actionsInProgress.size / Math.min(indexed, this.MAX_CONCURRENCY || DEFAULT_INDEXING_CONCURRENCY);
                            delayMs = Math.min(Math.max(delayMs, 1) * (desiredConcurrentRatio + Math.sqrt(indexed)) / (Math.sqrt(indexed) + 1), (actionsInProgress.size + 4) * 100);
                            if (this.isRetrying) {
                                await delay(1000);
                                delayMs = (delayMs + 10) * 2;
                            }
                            this.delayMs = delayMs;
                            lastTime = now + delayMs;
                            let completion = this.forQueueEntry(id);
                            if (completion && completion.then) {
                                completion.id = id;
                                actionsInProgress.add(completion);
                                completion.then(() => actionsInProgress.delete(completion));
                            }
                            if (sinceLastStateUpdate > (this.MAX_CONCURRENCY || DEFAULT_INDEXING_CONCURRENCY)) {
                                // we have process enough, commit our changes so far
                                this.averageConcurrencyLevel = ((this.averageConcurrencyLevel || 0) + actionsInProgress.size) / 2;
                                if (this.resumeFromKey) { // only update if we are actually resuming
                                    for (let last in actionsInProgress) {
                                        this.lastIndexingId = last.id;
                                        break;
                                    }
                                    this.resumeFromKey = this.lastIndexingId;
                                }
                            }
                            await delay(delayMs * desiredConcurrentRatio);
                        }
                        this.state = 'awaiting final indexing of ' + actionsInProgress.size;
                        await Promise.all(actionsInProgress); // then wait for all indexing to finish everything
                    } while (queue.size > 0);
                    await this.whenValueCommitted;
                    this.lastIndexedVersion = this.highestVersionToIndex;
                    if (initialQueueSize > 100 || initialQueueSize == undefined) {
                        console.log('Finished indexing', initialQueueSize || '', 'for', this.name);
                    }
                }
                catch (error) {
                    console.warn('Error occurred in processing index queue for', this.name, error);
                }
                this.state = 'processed';
                if (this.onStateChange) {
                    this.onStateChange({ processing: true, started: false });
                }
            }
            static forQueueEntry(id) {
                return this.tryForQueueEntry(id, () => when(this.get(id, INDEXING_MODE), () => {
                    let entry = this.db.cache.get(id);
                    return when(entry && entry.whenIndexed, () => {
                        if (this.queue)
                            this.queue.delete(id);
                    });
                }));
            }
            static requestProcessing(nice, queue) {
                // Indexing is performed one index at a time, until the indexing on that index is completed.
                // This is to prevent too much processing being consumed by the index processing,
                // and to allow dependent childStores to fully complete before downstream childStores start to
                // avoid thrashing from repeated changes in values
                if (this.whenProcessingThisComplete && !queue) {
                    // TODO: priority increases need to be transitively applied
                    this.nice = Math.min(this.nice, nice); // once started, niceness can only go down (and priority up)
                }
                else {
                    this.nice = nice;
                    this.state = 'pending';
                    this.whenProcessingThisComplete = Promise.all((this.sources || []).map(Source => Source.whenProcessingComplete)).then(() => this.processQueue(queue)).then(() => {
                        this.state = 'ready';
                        this.whenProcessingThisComplete = null;
                        //for (const sourceName in processingSourceVersions) {
                        //	sourceVersions[sourceName] = processingSourceVersions[sourceName]
                        //}
                        /*const event = new IndexingCompletionEvent()
                        event.sourceVersions = sourceVersions
                        event.sourceVersions[this.name] = lastIndexedVersion
                        super.updated(event, this)*/
                    });
                    this.whenProcessingThisComplete.queue = queue;
                    //this.whenProcessingThisComplete.version = lastIndexedVersion
                }
                return this.whenProcessingThisComplete;
            }
            static openDB(store, options) {
                let db = store.db = this.rootDB.openDB(store.dbName || store.name, Object.assign({
                    compression: true,
                    useFloat32: 3,
                    sharedStructuresKey: SHARED_STRUCTURE_KEY,
                }, options));
                store.rootDB = this.rootDB;
                return db;
            }
            static openChildDB(store, options) {
                let db = this.openDB(store, options);
                let rootStore = store.rootStore = this.rootStore || this;
                store.otherProcesses = rootStore.otherProcesses;
                let index;
                if (!rootStore.childStores) {
                    rootStore.childStores = [];
                }
                if (!this.indices) {
                    this.indices = [];
                }
                this.indices.push(store);
                rootStore.childStores.find((entry, i) => {
                    if (entry.name == store.name) {
                        index = i;
                        store.dbVersion = entry.dbVersion;
                        return true;
                    }
                });
                if (index > -1) {
                    store.dbVersion = rootStore.childStores[index].dbVersion;
                    rootStore.childStores[index] = store;
                }
                else {
                    rootStore.childStores.push(store);
                }
                return store.db;
            }
            static removeUnusedDBs() {
                let unusedDBs = new Set();
                for (let key of this.rootDB.getKeys({
                    start: Buffer.from([2])
                })) {
                    unusedDBs.add(key.toString());
                }
                for (let store of [this, ...(this.childStores || [])]) {
                    unusedDBs.delete(store.dbName || store.name);
                }
                console.log('Removing unused dbs', Array.from(unusedDBs));
            }
            static async resumeQueue() {
                this.resumeFromKey = this.db.get(INITIALIZING_LAST_KEY);
                if (!this.resumeFromKey) {
                    this.state = 'ready';
                    this.resumePromise = undefined;
                    return;
                }
                console.debug(this.name + ' Resuming from key ' + this.resumeFromKey);
                let idsToInitiallyIndex = this.getIdsFromKey(this.resumeFromKey);
                let db = this.db;
                const beforeCommit = () => {
                    if (this.resumeFromKey)
                        db.put(INITIALIZING_LAST_KEY, this.resumeFromKey, 1);
                };
                db.on('beforecommit', beforeCommit);
                this.state = 'building';
                console.debug('Created queue for initial index build', this.name);
                this.initialIndexCount = 0;
                await this.requestProcessing(30, idsToInitiallyIndex.map(id => {
                    this.initialIndexCount++;
                    return [id];
                }));
                console.debug('Finished initial index build of', this.name);
                db.off('beforecommit', beforeCommit);
                this.resumeFromKey = null;
                await db.remove(INITIALIZING_LAST_KEY);
                this.state = 'ready';
                this.resumePromise = undefined;
            }
            static resetAll() {
            }
        },
        _a.useWeakMap = true,
        _a.dbFolder = 'cachedb',
        _a));
};
const KeyValued = (Base, { versionProperty, valueProperty }) => { var _a; return _a = class extends Base {
        static get(id, mode) {
            let context = getCurrentContext();
            let entry = this.db.getEntry(id, mode ? NO_CACHE : 0);
            if (entry) {
                if (context) {
                    context.setVersion(entry.version);
                    if (context.ifModifiedSince >= entry.version) {
                        return NOT_MODIFIED;
                    }
                }
            }
            else {
                if (context) {
                    let version = getNextVersion();
                    context.setVersion(version);
                }
            }
            return entry && entry.value;
        }
        static is(id, value, event) {
            let entry = this.db.getEntry(id, NO_CACHE);
            if (!event) {
                event = entry ? new ReplacedEvent() : new AddedEvent();
            }
            event.triggers = [DISCOVERED_SOURCE];
            event.source = { constructor: this, id };
            event.version = getNextVersion();
            this.updated(event, { id, invalidate: false });
            if (entry) {
                entry.value = value;
            }
            else {
                entry = {
                    value,
                };
            }
            entry.version = event.version;
            return this.saveValue(id, entry, event.version);
        }
        static saveValue(id, entry, version) {
            this.highestVersionToIndex = Math.max(this.highestVersionToIndex || 0, version);
            let forValueResults = (this.indices && !entry.noIndex) ? this.indices.map(store => store.forValue(id, entry)) : [];
            let promises = forValueResults.filter(promise => promise && promise.then);
            const readyToCommit = (forValueResults) => {
                if (entry.version !== version)
                    return;
                entry.abortables = null; // indicate that this is no longer in process
                let conditionalVersion = entry.fromVersion;
                let committed = this.whenValueCommitted = this.db.ifVersion(id, conditionalVersion, () => {
                    // the whole set of writes for this entry and downstream indices are committed one transaction, conditional on the previous version
                    for (let result of forValueResults) {
                        if (result)
                            result.commit();
                    }
                    let committed;
                    let value = entry.value;
                    //console.log('conditional header for writing transform ' + (value ? 'write' : 'delete'), id, this.name, conditionalVersion)
                    if (value === undefined) {
                        if (conditionalVersion === null) {
                            // already an undefined entry, nothing to do (but clear out the transition)
                            if (this.db.cache.getEntry(id) == entry && entry.version === version) {
                                this.db.cache.delete(id);
                                return;
                            }
                        }
                        else {
                            this.whenWritten = committed = this.db.remove(id);
                        }
                    }
                    else {
                        this.whenWritten = committed = this.db.put(id, value, version);
                    }
                });
                return committed.then((successfulWrite) => {
                    //if (this.transitions.get(id) == transition && !transition.invalidating)
                    //	this.transitions.delete(id)
                    if (!successfulWrite) {
                        this.db.cache.delete(id);
                        let entry = this.db.getEntry(id);
                        console.debug('unsuccessful write of transform, data changed, updating', id, this.name, version, conditionalVersion, entry && entry.version);
                        let event = new ReloadEntryEvent();
                        if (entry && entry.version >= version) {
                            event.version = entry.version;
                        }
                        else {
                            event.doUpdate = true;
                            event.version = version;
                        }
                        this.updated(event, { id, constructor: this });
                    }
                });
            };
            let whenIndexedAndCommitted;
            if (promises.length == 0)
                whenIndexedAndCommitted = readyToCommit(forValueResults);
            else
                whenIndexedAndCommitted = (entry.whenIndexed = Promise.all(forValueResults)).then(readyToCommit);
            entry.committed = whenIndexedAndCommitted;
            return whenIndexedAndCommitted;
        }
        static getInstanceIds(range) {
            let db = this.db;
            range = range || {};
            range.start = range.start || true;
            range.values = false;
            if (range.waitForAllIds && this.ready) {
                delete range.waitForAllIds;
                return when(this.ready, () => this.getInstanceIds(range));
            }
            let iterable = db.getRange(range);
            return iterable;
        }
        static entries(range) {
            let db = this.db;
            return when(this.ready, () => {
                let results = db.getRange(Object.assign({
                    start: true,
                    versions: true,
                }, range));
                return range && range.asIterable ? results : results.asArray;
            });
        }
        /**
        * Iterate through all instances to find instances since the given version
        **/
        static getIdsFromKey(startKey) {
            //console.log('getInstanceIdsAndVersionsSince', this.name, sinceVersion)
            return this.db.getRange({
                start: startKey,
                snapshot: false,
                values: false,
            });
        }
        static remove(id, event) {
            if (id > 0 && typeof id === 'string' || !id) {
                throw new Error('Id should be a number or non-numeric string: ' + id);
            }
            return this.updated(event || (event = new DeletedEvent()), { id }).whenWritten;
        }
        setValue(value) {
            this.constructor.is(this.id, value);
        }
    },
    _a.reads = 0,
    _a.cachedReads = 0,
    _a; };
const PersistedBase = KeyValued(MakePersisted(Variable), {
    valueProperty: 'value',
    versionProperty: 'version'
});
class Persisted extends PersistedBase {
    static clearAllData() {
    }
    static set(id, value, event) {
        return this.is(id, value, event);
    }
    patch(properties) {
        return this.then((value) => {
            return when(this.put(value = Object.assign({}, value, properties)), () => value);
        });
    }
    put(value, event) {
        return this.constructor.is(this.id, value, event);
    }
}
Persisted.dbFolder = 'db';
Persisted.syncVersion = 10;
const Persistable = MakePersisted(Transform);
class Cached extends KeyValued(MakePersisted(Transform), {
    valueProperty: 'cachedValue',
    versionProperty: 'cachedVersion'
}) {
    static get(id, mode) {
        let context = getCurrentContext();
        return when(this.whenUpdatedInContext(), () => {
            let entry = this.db.getEntry(id, mode ? NO_CACHE : 0);
            if (entry) {
                if (!entry.value) { // or only undefined?
                    let abortables = entry.abortables;
                    //console.log('Running transform on invalidated', id, this.name, this.createHeader(entry[VERSION]), oldTransition)
                    this.runTransform(id, entry, mode);
                    /*if (abortables) {
                        // if it is still in progress, we can abort it and replace the result
                        //oldTransition.replaceWith = transition.value
                        for (let abortable of abortables) {
                            abortable()
                        }
                    }*/
                    return entry.value;
                }
                if (context) {
                    context.setVersion(entry.version);
                    if (context.ifModifiedSince >= entry.version) {
                        return NOT_MODIFIED;
                    }
                }
                when(entry.value, (result) => deepFreeze);
                return entry.value;
            }
            let version = getNextVersion();
            if (context)
                context.setVersion(version);
            entry = { version };
            entry = this.runTransform(id, entry, mode);
            when(entry.value, (result) => {
                if (result !== undefined && !entry.invalidating) {
                    deepFreeze(result, 0);
                    let event = new DiscoveredEvent();
                    event.triggers = [DISCOVERED_SOURCE];
                    event.source = { constructor: this, id };
                    event.version = version;
                    this.instanceSetUpdated(event);
                    this.updated(event, {
                        id,
                        constructor: this
                    });
                }
            });
            return entry.value;
        });
    }
    static runTransform(id, entry, mode) {
        let version = entry.version;
        if (!entry.abortables)
            entry.abortables = [];
        /*entry = {
            version,
            previousValue: entry.previousValue,
            abortables: []
        }*/
        let cache = this.db.cache;
        cache.set(id, entry, -1); // enter in cache without LRFU tracking, keeping it in memory, this should be entered into LRFU once it is committed by the lmdb-store caching store logic
        const removeTransition = () => {
            if (cache.get(id) === entry && !entry.invalidating)
                cache.delete(id);
        };
        let hasPromises;
        let inputData = this.sources ? this.sources.map(source => {
            let data = source.get(id, AS_SOURCE);
            if (data && data.then) {
                hasPromises = true;
            }
            return data;
        }) : [];
        try {
            entry.value = when(when(hasPromises ? Promise.all(inputData) : inputData, inputData => {
                if (inputData.length > 0 && inputData[0] === undefined && !this.sourceOptional) // first input is undefined, we pass through
                    return;
                let context = getCurrentContext();
                let transformContext = context ? context.newContext() : new RequestContext(null, null);
                transformContext.abortables = entry.abortables;
                inputData.push(id);
                return this.prototype.transform ? transformContext.executeWithin(() => this.prototype.transform.apply({ id }, inputData)) : inputData;
            }), result => {
                if (entry.version != version) {
                    if (entry.replaceWith) {
                        return entry.replaceWith;
                    }
                    return result;
                } // else normal transform path
                entry.value = result;
                this.saveValue(id, entry, version);
                return result;
            }, (error) => {
                removeTransition();
                if (error.__CANCEL__) {
                    return entry.replaceWith;
                }
                throw error;
            });
        }
        catch (error) {
            removeTransition();
            throw error;
        }
        return entry;
    }
    getValue() {
        return this.constructor.get(this.id);
    }
    is(value, event) {
        // we skip getEntryData and pretend it wasn't in the cache... not clear if
        // that is how we want is() to behave or not
        this.constructor.is(this.id, value, event);
        return this;
    }
    static openChildDB(store, options) {
        if (!this.queue) {
            this.queue = new Map();
        }
        return super.openChildDB(store, options);
    }
    static updated(event, by) {
        let id = by && (typeof by === 'object' ? by.id : by); // if we are getting an update from a source instance
        event = super.updated(event, by);
        if (this.queue) {
            if (by && by.constructor === this || // our own instances can notify us of events, ignore them
                this.otherProcesses && event.sourceProcess &&
                    !(id && this.queue.has(id)) && // if it is in our queue, we need to update the version number in our queue
                    (this.otherProcesses.includes(event.sourceProcess) || // another process should be able to handle this
                        this.otherProcesses.some(otherProcessId => otherProcessId < process.pid))) { // otherwise, defer to the lowest number process to handle it
                // we can skip these (unless they are in are queue, then we need to update)
                return event;
            }
            if (id && (event.type === 'discovered')) {
                this.enqueue(id, event);
            }
        }
        return event;
    }
    static enqueue(id, event, previousEntry) {
        if (this.resumeFromKey && lmdb.compareKey(this.resumeFromKey, id) < 0) // during initialization, we ignore updates because we are going rebuild
            return;
        const version = event.version;
        // queue up processing the event
        let indexRequest = this.queue.get(id);
        if (indexRequest) {
            // put it at that end so version numbers are in order, but don't alter the previous state or version, that is still what we will be diffing from
            this.queue.delete(id);
            this.queue.set(id, indexRequest);
            indexRequest.version = version;
            if (event.triggers)
                for (let trigger of event.triggers)
                    indexRequest.triggers.add(trigger);
        }
        else {
            this.queue.set(id, indexRequest = {
                version: version,
                previousEntry,
                now: Date.now(),
                triggers: event.triggers instanceof Set ? event.triggers : new Set(event.triggers),
            });
            /*if (indexRequest.previousState == INVALIDATED_ENTRY) {
                indexRequest.fromValues = event.fromValues // need to have a shared map to update
                indexRequest.by = by
            }*/
            this.requestProcessing(DEFAULT_INDEXING_DELAY);
        }
        if (!version) {
            throw new Error('missing version');
        }
        indexRequest.deleted = event.type == 'deleted';
    }
    static async resetAll() {
        console.debug('reseting', this.name);
        let version = this.startVersion = 1;
        let allIds = this.fetchAllIds ? await this.fetchAllIds() :
            (this.sources && this.sources[0] && this.sources[0].getInstanceIds) ?
                await this.sources[0].getInstanceIds({
                    waitForAllIds: true,
                }) : [];
        let committed;
        let queued = 0;
        console.log('loading ids for', this.name, 'with', allIds.length, 'ids');
        let idCount = 0;
        for (let id of allIds) {
            idCount++;
            if (this.instancesById.getValue(id)) {
                // instance already in memory
                this.for(id).updated();
                continue;
            }
            this.lastVersion = version++; // we give each entry its own version so that downstream childStores have unique versions to go off of
            this.whenWritten = committed = this.db.put(id, 0, -version);
            if (queued++ > 2000) {
                console.log('writing block of ids');
                await this.whenWritten;
                console.log('wrote block of ids');
                queued = 0;
            }
        }
        console.log('loaded ids for', this.name, 'with', idCount, 'ids');
        return committed;
    }
    static invalidateEntry(id, event, by) {
        let db = this.db;
        let written;
        if (event && event.sourceProcess) {
            // if it came from another process we can count on it to have written the update
            db.cache.delete(id); // clear from cache
            return;
        }
        let version = event.version;
        if (this.indices) {
            let entry = db.getEntry(id);
            if (entry) {
                db.cache.expirer.used(entry, -1); // key it pinned in memory
                if (!entry.abortables) { // if this entry is in a transform and not committed, don't update fromVersion
                    entry.previousValue = entry.value;
                    entry.fromVersion = entry.version;
                    entry.abortables = [];
                }
                entry.value = null; // set as invalidated
                entry.version = version; // new version
                // for deleted events, let the transform do the removal
            }
            else {
                entry = {
                    version,
                    abortables: [],
                };
                db.cache.set(id, entry, -1); // enter in cache without LRFU tracking, keeping it in memory
            }
            if (event.noIndex)
                entry.noIndex = true;
            if (!by || by.invalidate !== false) {
                written = this.forQueueEntry(id).then(() => entry.committed);
                let lastCompletion = this.whenIndexedAndCommitted = (this.whenIndexedAndCommitted ?
                    Promise.all([this.whenIndexedAndCommitted, written]) : written).then(() => {
                    if (this.whenIndexedAndCommitted == lastCompletion)
                        this.whenIndexedAndCommitted = null;
                });
            }
        }
        else {
            if (event && event.type === 'deleted') {
                // completely empty entry for deleted items
                written = db.remove(id);
            }
            else if (!by || by.invalidate !== false) {
                written = db.put(id, null, version);
            }
        }
        this.whenWritten = written;
        if (!event.whenWritten)
            event.whenWritten = written;
    }
    static async receiveRequest({ id, waitFor }) {
        if (waitFor == 'get') {
            console.log('waiting for entity to commit', this.name, id);
            this.clearEntryCache(id);
            await this.get(id);
            let entry = this.db.getEntry(id);
            if (entry) {
                // wait for get to truly finish and be committed
                await entry.whenIndexed;
                await entry.committed;
            }
            // return nothing, so we don't create any overhead between processes
        }
    }
    static get version() {
        if (this.sources) {
            return this.sources.reduce((sum, Source) => sum += (Source.version || 0), this._version);
        }
        else {
            return this._version || 1;
        }
    }
    static set version(version) {
        this._version = version;
    }
    static from(...sources) {
        if (!sources[0]) {
            throw new Error('No source provided');
        }
        let Cached = class extends this {
            get checkSourceVersions() {
                return false;
            }
        };
        for (let Source of sources) {
            if (Source.returnsAsyncIterables) {
                this.returnsAsyncIterables;
            }
        }
        Cached.sources = sources;
        return Cached;
    }
    static derivedFrom(...sources) {
        for (let source of sources) {
            if (source.notifies) {
                if (!this.sources)
                    this.sources = [];
                this.sources.push(source);
            }
            else if (typeof source === 'function') {
                this.prototype.transform = source;
            }
            else {
                Object.assign(this, source);
            }
        }
        this.start();
    }
    static getInstanceIds(range) {
        if (!this.fetchAllIds && this.sources && this.sources[0] && this.sources[0].getInstanceIds) {
            // if we don't know if we have all our ids, our source is a more reliable source of instance ids
            return this.sources[0].getInstanceIds(range);
        }
        return super.getInstanceIds(range);
    }
    static updateDBVersion() {
        if (this.indices) {
            console.debug('Setting up indexing for', this.name);
            this.resumeFromKey = true;
            this.db.putSync(INITIALIZING_LAST_KEY, true);
        }
        super.updateDBVersion();
    }
    static get whenProcessingComplete() {
        return this.sources && Promise.all(this.sources.map(Source => Source.whenProcessingComplete));
    }
}
function secureAccess(Class) {
    Class.allow = function (...permissions) {
        let Class = this;
        let methodOverrides = {
            for(id) {
                let target = Class.for(id);
                return new Proxy(target, handler);
            },
            stopNotifies(target) {
                // skip permissions on this
                return this.stopNotifies(target);
            },
            isChecked() {
                return true;
            }
        };
        let handler = {
            get(target, name) {
                let value = target[name];
                if (methodOverrides[name]) {
                    return methodOverrides[name].bind(target);
                }
                if (typeof value === 'function') {
                    return function () {
                        let context = getCurrentContext();
                        // create a new derivative context that includes the session, but won't
                        // update the version/timestamp
                        return context.newContext().executeWithin(() => {
                            const permitted = when(secureAccess.checkPermissions(permissions, target, name, Array.from(arguments)), (permitted) => {
                                if (permitted !== true) {
                                    throw new AccessError('User does not have required permissions: ' + permitted + ' for ' + Class.name);
                                }
                            });
                            const whenPermitted = () => context.executeWithin(() => value.apply(target, arguments));
                            if (permitted.then) {
                                let result;
                                let whenFinished = permitted.then(() => {
                                    result = whenPermitted();
                                });
                                return {
                                    then: (onFulfilled, onRejected) => whenFinished.then(() => {
                                        return onFulfilled(result);
                                    }, onRejected)
                                };
                            }
                            return whenPermitted();
                        });
                    };
                }
                else {
                    return value;
                }
            }
        };
        return new Proxy(this, handler);
    };
    return Class;
}
class DiscoveredEvent extends AddedEvent {
}
DiscoveredEvent.prototype.type = 'discovered';
class ReloadEntryEvent extends ReplacedEvent {
}
ReloadEntryEvent.prototype.type = 'reload-entry';
function getCurrentStatus() {
    return Array.from(allStores.values()).map(store => ({
        name: store.name,
        indexed: store.initialIndexCount,
        queueSize: store.queue && store.queue.size,
        size: global.skipDBStats ? 0 : store.db.getStats().entryCount,
        state: store.state,
        concurrencyLevel: store.actionsInProgress ? store.actionsInProgress.size : 0,
        //pendingRequests: Array.from(Index.pendingRequests),
    }));
}
secureAccess.checkPermissions = () => true;
let clearOnStart;
let sharedStructureDirectory;
let verboseLogging;
let storesObject = global;
function configure(options) {
    Persisted.dbFolder = options.dbFolder;
    Cached.dbFolder = options.cacheDbFolder || options.dbFolder;
    Persistable.dbFolder = options.cacheDbFolder || options.dbFolder;
    globalDoesInitialization = options.doesInitialization;
    verboseLogging = options.verboseLogging;
    clearOnStart = options.clearOnStart;
    if (options.storesObject) {
        storesObject = options.storesObject;
    }
    if (options.getCurrentContext)
        getCurrentContext = options.getCurrentContext;
    if (options.sharedStructureDirectory)
        sharedStructureDirectory = options.sharedStructureDirectory;
    if (options.sharedInstrumenting) {
        console.warn('sharedInstrumenting is turned on!!!!!!!');
    }
}
function writeCommonStructures() {
    let wrote = [];
    for (let [name, store] of allStores) {
        if (store.writeCommonStructure())
            wrote.push(name);
    }
    return wrote;
}
const delay = ms => new Promise(resolve => ms >= 1 ? setTimeout(resolve, ms) : setImmediate(resolve));
let deepFreeze = process.env.NODE_ENV == 'development' ? (object, depth) => {
    if (depth > 100)
        throw new Error('Max object depth exceeded or circular reference in data');
    if (object && typeof object == 'object') {
        if (object.constructor == Object) {
            for (let key in object) {
                let value = object[key];
                if (typeof value == 'object')
                    deepFreeze(value, (depth || 0) + 1);
            }
        }
        else if (object.constructor == Array) {
            for (let i = 0, l = object.length; i < l; i++) {
                let value = object[i];
                if (typeof value == 'object')
                    deepFreeze(value, (depth || 0) + 1);
            }
        }
    }
    return object;
} : (object) => object;

const { VArray: VArray$1, ReplacedEvent: ReplacedEvent$1, UpdateEvent: UpdateEvent$2, getNextVersion: getNextVersion$1 } = alkali.VArray ? alkali : alkali__default;
//import { mergeProgress, registerProcessing, whenClassIsReady, DEFAULT_CONTEXT } from './UpdateProgress'
const expirationStrategy$1 = ExpirationStrategy.defaultInstance;
const INITIALIZING_LAST_KEY$1 = Buffer.from([1, 7]);
const INDEXING_MODE$1 = { indexing: true };
const INITIALIZATION_SOURCE = 'is-initializing';
const QUEUED_UPDATES = Symbol.for('qu');
const Index = ({ Source }) => {
    const Index = class extends KeyIndex {
    };
    Source.updateWithPrevious = true;
    Index.sources = [Index.source = Source];
    Index.eventLog = [];
    Index.queuedUpdateArrays = [];
    Index.allQueuedUpdates = new Set();
    return Index;
};
class KeyIndex extends Persistable.as(VArray$1) {
    static derivedFrom(...sources) {
        for (let source of sources) {
            if (source.notifies) {
                if (!this.sources)
                    this.sources = [];
                this.sources.push(this.source = source);
                source.updateWithPrevious = true;
            }
            else if (typeof source === 'function') {
                this.indexBy = source;
            }
            else {
                Object.assign(this, source);
            }
        }
        /*		static whenProcessingComplete: Promise<any> // promise for the completion of processing in current indexing task for this index,
                whenCommitted: Promise<any> // promise for when an update received by this index has been fully committed (to disk),
                indexingProcess: Promise<any>*/
        this.eventLog = [];
        this.queuedUpdateArrays = [];
        this.allQueuedUpdates = new Set();
        this.start();
    }
    static forValue(id, transition) {
        return this.tryForQueueEntry(id, () => this.indexEntry(id, transition));
    }
    static forQueueEntry(id) {
        return this.tryForQueueEntry(id, () => this.indexEntry(id).then(complete => {
            if (complete) {
                complete.commit();
            }
        }));
    }
    static async indexEntry(id, transition) {
        let { previousValue: previousData, deleted, sources, triggers, fromVersion: previousVersion } = transition || {};
        let operations = [];
        let eventUpdateSources = [];
        let toRemove = new Map();
        // TODO: handle delta, for optimized index updaes
        // this is for recording changed entities and removing the values that previously had been indexed
        let previousEntries;
        try {
            if (previousData != null) { // if no data, then presumably no references to clear
                // use the same mapping function to determine values to remove
                previousEntries = previousData === undefined ? previousData : this.indexBy(previousData, id);
                if (previousEntries && previousEntries.then)
                    previousEntries = await previousEntries;
                if (typeof previousEntries == 'object' && previousEntries) {
                    previousEntries = this.normalizeEntries(previousEntries);
                    for (let entry of previousEntries) {
                        let previousValue = entry.value;
                        previousValue = this.db.encoder.encode(previousValue);
                        toRemove.set(typeof entry === 'object' ? entry.key : entry, previousValue);
                    }
                }
                else if (previousEntries != null) {
                    toRemove.set(previousEntries, this.db.encoder.encode(previousEntries));
                }
            }
        }
        catch (error) {
            if (error.isTemporary)
                throw error;
            //if (transition && transition.version !== version) return // don't log errors from invalidated states
            this.warn('Error indexing previous value', this.source.name, 'for', this.name, id, error);
        }
        //if (indexRequest && indexRequest.version !== version) return // if at any point it is invalidated, break out
        let entries;
        if (!deleted) {
            let data;
            try {
                data = transition ? transition.value : this.source.get(id, INDEXING_MODE$1);
                if (data && data.then)
                    data = await data;
            }
            catch (error) {
                if (error.isTemporary)
                    throw error;
                try {
                    // try again
                    data = transition ? transition.value : await this.source.get(id, INDEXING_MODE$1);
                }
                catch (error) {
                    //if (indexRequest && indexRequest.version !== version) return // if at any point it is invalidated, break out
                    this.warn('Error retrieving value needing to be indexed', error, 'for', this.name, id);
                    data = undefined;
                }
            }
            //if (indexRequest && indexRequest.version !== version) return // if at any point it is invalidated, break out
            // let the indexBy define how we get the set of values to index
            try {
                entries = data === undefined ? data : this.indexBy(data, id);
                if (entries && entries.then)
                    entries = await entries;
            }
            catch (error) {
                if (error.isTemporary)
                    throw error;
                //if (indexRequest && indexRequest.version !== version) return // if at any point it is invalidated, break out
                this.warn('Error indexing value', error, 'for', this.name, id);
                entries = undefined;
            }
            entries = this.normalizeEntries(entries);
            for (let entry of entries) {
                // we use the composite key, so we can quickly traverse all the entries under a certain key
                let key = typeof entry === 'object' ? entry.key : entry; // TODO: Maybe at some point we support dates as keys
                // TODO: If toRemove has the key, that means the key exists, and we don't need to do anything, as long as the value matches (if there is no value might be a reasonable check)
                let removedValue = toRemove.get(key);
                // a value of '' is treated as a reference to the source object, so should always be treated as a change
                let value = this.db.encoder.encode(entry.value);
                if (removedValue != null)
                    toRemove.delete(key);
                let isChanged = removedValue == null || !value.equals(removedValue);
                if (isChanged || entry.value == null || this.alwaysUpdate) {
                    if (isChanged) {
                        let fullKey = [key, id];
                        operations.push({
                            key: fullKey,
                            value: value
                        });
                        operations.byteCount = (operations.byteCount || 0) + value.length + fullKey.length;
                    }
                    if (!this.resumePromise && this.listeners && this.listeners.length > 0)
                        eventUpdateSources.push({ key, sources, triggers });
                }
            }
        }
        for (let [key] of toRemove) {
            /*if (fullKey.length > 510) {
                console.error('Too large of key indexing', this.name, key)
                continue
            }*/
            operations.push({
                key: [key, id]
            });
            if (!this.resumePromise && this.listeners && this.listeners.length > 0)
                eventUpdateSources.push({ key, sources, triggers });
        }
        if (Index.onIndexEntry) {
            Index.onIndexEntry(this.name, id, version, previousEntries, entries);
        }
        return {
            commit: () => {
                let batchFinished;
                let encoder = this.db.encoder;
                this.db.encoder = null;
                try {
                    for (let operation of operations) {
                        if (operation.value)
                            batchFinished = this.db.put(operation.key, operation.value);
                        else
                            batchFinished = this.db.remove(operation.key);
                    }
                }
                finally {
                    this.db.encoder = encoder;
                }
                if (eventUpdateSources.length > 0) {
                    return (batchFinished || Promise.resolve()).then(() => this.queueUpdates(eventUpdateSources));
                }
            }
        };
    }
    static normalizeEntries(entries) {
        if (typeof entries != 'object') {
            // allow single primitive key
            return entries == null ? [] : [entries];
        }
        else if (entries instanceof Map) {
            return Array.from(entries).map(([key, value]) => ({ key, value }));
        }
        else if (!(entries instanceof Array)) {
            // single object
            if (entries === null)
                return [];
            return [entries];
        }
        return entries;
    }
    static log(...args) {
        this.eventLog.push(args.join(' ') + ' ' + new Date().toLocaleString());
        console.log(...args);
    }
    static warn(...args) {
        this.eventLog.push(args.join(' ') + ' ' + new Date().toLocaleString());
        console.warn(...args);
    }
    static updated(event, by) {
        // don't do anything, we don't want these events to propagate through here, and we do indexing based on upstream queue
    }
    static queueUpdates(eventSources) {
        if (!this.unsavedUpdates) {
            this.unsavedUpdates = [];
            this.unsavedUpdates.id = this.unsavedQueuedId = (this.unsavedQueuedId || 0) + 1;
            if (this.queuedUpdateArrays.length == 0)
                setTimeout(() => this.processQueuedUpdates());
            this.queuedUpdateArrays.push(this.unsavedUpdates);
        }
        for (const { key, triggers, sources } of eventSources) {
            if (!this.allQueuedUpdates.has(key)) {
                this.allQueuedUpdates.add(key);
                this.unsavedUpdates.push(key);
            }
        }
    }
    static async processQueuedUpdates() {
        let inProgress = [];
        while (this.queuedUpdateArrays.length) {
            let updateArray = this.queuedUpdateArrays[0];
            let l = updateArray.length;
            for (let i = updateArray.sent || 0; i < l; i++) {
                let key = updateArray[i];
                try {
                    let event = new ReplacedEvent$1();
                    this.allQueuedUpdates.delete(key);
                    let start = process.hrtime.bigint();
                    super.updated(event, {
                        id: key,
                        constructor: this
                    });
                    let whenWritten = event.whenWritten;
                    if (whenWritten) {
                        inProgress.push(event.whenWritten);
                        if (inProgress.length > 20) {
                            await Promise.all(inProgress);
                            inProgress = [];
                        }
                    }
                    await delay$1(Number(process.hrtime.bigint() - start) / 1000000); // delay about the same amount of time the update took
                }
                catch (error) {
                    this.warn('Error sending index updates', error);
                }
            }
            if (updateArray.written) {
                Promise.all(inProgress).then(() => this.db.remove([QUEUED_UPDATES, updateArray.id]));
                this.queuedUpdateArrays.shift();
            }
            else {
                if (l == updateArray.length) {
                    await Promise.all(inProgress);
                    inProgress = [];
                }
                if (l == updateArray.length && this.unsavedUpdates == updateArray) {
                    this.unsavedUpdates = null; // we sent all the entries before it was written, restart a new array
                    this.queuedUpdateArrays.shift();
                }
                else
                    updateArray.sent = l;
            }
        }
    }
    static sendUpdates(eventSources) {
        let updatedIndexEntries = new Map();
        // aggregate them by key so as to minimize the number of events we send
        nextEvent: for (const { key, triggers, sources } of eventSources) {
            let entry = updatedIndexEntries.get(key);
            if (!entry) {
                updatedIndexEntries.set(key, entry = {
                    sources: new Set(),
                    triggers: new Set(),
                });
            }
            if (triggers) {
                for (let trigger of triggers) {
                    entry.triggers.add(trigger);
                    if (trigger === INITIALIZATION_SOURCE) {
                        continue nextEvent; // don't record sources for initialization
                    }
                }
            }
            if (sources)
                for (let source of sources)
                    entry.sources.add(source);
        }
        let updatedIndexEntriesArray = Array.from(updatedIndexEntries).reverse();
        updatedIndexEntries = new Map();
        let indexedEntry;
        while ((indexedEntry = updatedIndexEntriesArray.pop())) {
            try {
                let event = new ReplacedEvent$1();
                let indexEntryUpdate = indexedEntry[1];
                event.sources = indexEntryUpdate.sources;
                event.triggers = Array.from(indexEntryUpdate.triggers);
                super.updated(event, {
                    id: indexedEntry[0],
                    constructor: this
                });
            }
            catch (error) {
                this.warn('Error sending index updates', error);
            }
        }
        this.instanceIds.updated();
    }
    static get(id) {
        // First: ensure that all the source instances are up-to-date
        return when(this.source.whenUpdatedInContext(true), () => {
            let iterable = this._getIndexedValues({
                start: id,
                end: [id, '\u{ffff}'],
            });
            return this.returnsIterables ? iterable : iterable.asArray;
        });
    }
    static getIndexedKeys(id) {
        return this._getIndexedValues({
            start: id,
            end: [id, '\u{ffff}'],
            values: false,
        }, true);
    }
    static getIndexedValues(range) {
        range = range || {};
        if (!this.initialized && range.waitForInitialization) {
            return this.start().then(() => this.getIndexedValues(range));
        }
        if (range.start === undefined)
            range.start = true;
        return when(!range.noWait && this.whenUpdatedInContext(), () => this._getIndexedValues(range, !range.onlyValues, range.useFullKey));
    }
    // Get a range of indexed entries for this id (used by Reduced)
    static _getIndexedValues(range, returnKeys, useFullKey) {
        const db = this.db;
        return db.getRange(range).map(({ key, value }) => {
            let [, sourceId] = key;
            /*if (range.recordApproximateSize) {
                let approximateSize = approximateSize += key.length + (value && value.length || 10)
            }*/
            let parsedValue = value == null ? this.source.get(sourceId) : value;
            if (parsedValue && parsedValue.then) {
                return parsedValue.then(parsedValue => returnKeys ? {
                    key: useFullKey ? key : sourceId,
                    value: parsedValue,
                } : parsedValue);
            }
            return returnKeys ? {
                key: useFullKey ? key : sourceId,
                value: parsedValue,
            } : parsedValue;
        });
    }
    /**
    * Indexing function, that defines the keys and values used in the indexed table.
    * This should be implemented by Index subclasses, and should be safe/functional
    * method with referential integrity (always returns the same results with same inputs),
    * as it is used to determine key/values on both addition and removal of entities.
    * @param data The object to be indexed
    * @return The return value can be an array of objects, where each object has a `key` and a `value`. It can only be an array of simple strings or numbers, if it is merely keys that need to be indexed, or even be a just a string (or number), if only a single key should be indexed
    **/
    static indexBy(data, sourceKey) {
        return null;
    }
    static whenUpdatedInContext(context) {
        return this.source.whenUpdatedInContext(true);
    }
    // static returnsIterables = true // maybe at some point default this to on
    static getInstanceIdsAndVersionsSince(version) {
        // There is no version tracking with indices.
        // however, indices always do send updates, and as long as we wait until we are ready and finished with initial indexing
        // downstream tables should have received all the updates they need to proceed
        //console.log('getInstanceIdsAndVersionsSince from KeyIndex', this.name, version)
        return this.ready.then(() => {
            //this.log('getInstanceIdsAndVersionsSince ready from KeyIndex', this.name, version)
            if (version == 0) { // if we are starting from scratch, we can return everything
                return when(this.getInstanceIds(), idsAndVersions => {
                    idsAndVersions = idsAndVersions.map(id => ({
                        id,
                        version: getNextVersion$1(),
                    }));
                    idsAndVersions.isFullReset = true;
                    return idsAndVersions;
                });
            }
            return [];
        });
    }
    clearCache() {
        this.cachedValue = undefined;
        this.cachedVersion = -1;
    }
    valueOf() {
        return when(super.valueOf(true), (value) => {
            expirationStrategy$1.useEntry(this, (this.approximateSize || 100) * 10); // multiply by 10 because generally we want to expire index values pretty quickly
            return value;
        });
    }
    static openDatabase() {
        let db = this.source.openChildDB(this);
        db.on('beforecommit', () => {
            if (this.unsavedUpdates && this.unsavedUpdates.length > 0) {
                this.unsavedUpdates.written = true;
                db.put([QUEUED_UPDATES, this.unsavedQueuedId], this.unsavedUpdates);
                this.unsavedUpdates = null;
            }
        });
        return false; // is not root
    }
    static async initializeData() {
        await super.initializeData();
        let hadQueuedUpdates = this.queuedUpdateArrays.length > 0;
        for (let { key, value } of this.db.getRange({
            start: [QUEUED_UPDATES, 0],
            end: [QUEUED_UPDATES, 1e30],
        })) {
            value.id = key[1];
            value.written = true;
            this.unsavedQueuedId = Math.max(this.unsavedQueuedId || 0, value.id);
            this.queuedUpdateArrays.push(value);
            for (let updateId of value) {
                this.allQueuedUpdates.add(updateId);
            }
        }
        if (this.queuedUpdateArrays.length > 0 && !hadQueuedUpdates)
            this.processQueuedUpdates();
    }
    static getIdsFromKey(key) {
        return this.source.getIdsFromKey(key);
    }
    static updateDBVersion() {
        if (!this.source.wasReset) { // only reindex if the source didn't do it for use
            this.resumeFromKey = true;
            this.db.putSync(INITIALIZING_LAST_KEY$1, true);
        }
        super.updateDBVersion();
    }
    static resumeQueue() {
        this.state = 'waiting for upstream source to build';
        // explicitly wait for source to finish resuming before our own resuming
        return when(this.source.resumePromise, () => super.resumeQueue());
    }
    static async clearEntries(set) {
        let result;
        let db = this.db;
        let i = 1;
        try {
            for (let key of db.getRange({
                start: true,
                values: false,
            })) {
                let [, sourceId] = key;
                if (set.has(sourceId)) {
                    result = db.removeSync(key);
                }
                if (i++ % 100000 == 0)
                    await delay$1();
            }
        }
        catch (error) {
            console.error(error);
        }
        return result; // just need to wait for last one to finish (guarantees all others are finished)
    }
    static loadVersions() {
        // don't load versions
    }
    resetCache() {
        // don't reset any in the db, we are incrementally updating
        this.cachedValue = undefined;
        this.updateVersion();
    }
    static get instances() {
        // don't load from disk
        return this._instances || (this._instances = []);
    }
    static getInstanceIds(range) {
        let db = this.db;
        range = range || {};
        if (range.start === undefined)
            range.start = true;
        let whenReady;
        if (range.waitForAllIds) {
            whenReady = when(this.ready, () => when(this.resumePromise, () => this.whenProcessingComplete));
        }
        let lastKey;
        range.values = false;
        return when(whenReady, () => db.getRange(range).map((key) => key[0]).filter(key => {
            if (key !== lastKey) { // skip multiple entries under one key
                lastKey = key;
                return true;
            }
        }));
    }
}
KeyIndex.eventLog = [];
KeyIndex.queuedUpdateArrays = [];
KeyIndex.pendingEvents = new Map();
KeyIndex.myEarliestPendingVersion = 0; // have we registered our process, and at what version
Index.from = (Source) => Index({ Source });
const delay$1 = (ms) => new Promise(resolve => ms >= 1 ? setTimeout(resolve, ms) : setImmediate(resolve));

const INVALIDATED_VALUE = Buffer.from([]);
const SEPARATOR_BYTE = Buffer.from([30]); // record separator control character
const REDUCED_INDEX_PREFIX_BYTE = Buffer.from([3]);
const CHILDREN = 2;
class Reduced extends Cached {
    /**
    * This defines the reduce function that accumulates index entry values into a single value
    */
    reduceBy(a, b) {
        return null;
    }
    static derivedFrom(...sources) {
        for (let source of sources) {
            if (source.notifies) {
                if (!this.sources)
                    this.sources = [];
                this.sources.push(source);
            }
            else if (typeof source === 'function') {
                this.reduceBy = source;
            }
            else {
                Object.assign(this, source);
            }
        }
        this.start();
    }
    // rewrite the source to be the computed reduced value
    // this allows transform to still execute using the result
    get source() {
        return this.reducingSource || (this.reducingSource = {
            valueOf: this.getReducedEntry.bind(this)
        });
    }
    set source(source) {
        this.indexSource = source;
    }
    getReducedEntry() {
        return this.transaction(async (db, put) => {
            if (this.rootLevel > -1) {
                const indexKey = toBufferKey(this.id);
                const { split, noChildren, accumulator } = await this.reduceRange(this.rootLevel, Buffer.from([1]), Buffer.from([255]), put);
                if (split) // splitting the root node, just bump up the level number
                    this.rootLevel++;
                else if (noChildren) {
                    // if all children go away, return to a root level of 1
                    // we don't ever incrementally reduce depth, and if we are decreasing children,
                    // we can represent a single childen with an arbitrarily deep single-child-at-every-level
                    // tree
                    this.rootLevel = 1;
                }
                // now it should be written to the node
                // this should be done by Cached: Class.dbPut(this.id, version + ',' + this.rootLevel + ',' + serialize(accumulator))
                return accumulator;
            }
        });
    }
    async reduceRange(level, rangeStartKey, rangeEndKey, put) {
        let iterator;
        const Class = this.constructor;
        const db = Class.db;
        const indexBufferKey = toBufferKey(this.id);
        if (level === 1) {
            // leaf-node, go to source index
            iterator = this.indexSource._getIndexedValues({
                start: Buffer.concat([indexBufferKey, SEPARATOR_BYTE, rangeStartKey]),
                end: Buffer.concat([indexBufferKey, SEPARATOR_BYTE, rangeEndKey]),
            }, true)[Symbol.iterator]();
        }
        else {
            // mid-node, use our own nodes/ranges here
            iterator = db.getRange({
                start: Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level - 1]), indexBufferKey, SEPARATOR_BYTE, rangeStartKey]),
                end: Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level - 1]), indexBufferKey, SEPARATOR_BYTE, rangeEndKey]),
                reverse: false,
            }).map(({ key, value }) => {
                let [, startKey, endKey] = fromBufferKey(key.slice(2), true);
                return {
                    level: key[1],
                    key: startKey,
                    endKey,
                    value: value.length > 0 ? parse(value) : INVALIDATED_VALUE,
                };
            })[Symbol.iterator]();
        }
        let next;
        let version = Date.now();
        let firstOfSection = true;
        let split = false;
        let lastDividingKey = rangeStartKey;
        let accumulator;
        let totalAccumulator;
        let childrenProcessed = 0; // asynchronously iterate
        while (!(next = iterator.next()).done) {
            let { key, endKey, value } = next.value;
            if (value && value.then) // if the index has references to variables, need to resolve them
                value = await value;
            childrenProcessed++;
            if (childrenProcessed > CHILDREN) {
                childrenProcessed = 0;
                let nextDividingKey = endKey || key;
                put(Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level]), indexBufferKey, SEPARATOR_BYTE, lastDividingKey, SEPARATOR_BYTE, lastDividingKey = toBufferKey(nextDividingKey)]), serialize(accumulator));
                if (!split)
                    totalAccumulator = accumulator; // start with existing accumulation
                else
                    totalAccumulator = this.reduceBy(totalAccumulator, accumulator);
                split = true;
                firstOfSection = true;
            }
            if (value == INVALIDATED_VALUE) {
                const result = await this.reduceRange(level - 1, toBufferKey(key), toBufferKey(endKey), put);
                value = result.accumulator;
                put(Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level - 1]), indexBufferKey, SEPARATOR_BYTE, toBufferKey(key), SEPARATOR_BYTE, toBufferKey(endKey)]), result.split || result.noChildren ?
                    undefined : // if it is a split, we have to remove the existing node
                    serialize(value)); // otherwise write our value
                if (result.noChildren) {
                    continue;
                }
            }
            if (firstOfSection) {
                accumulator = value;
            }
            else {
                accumulator = await this.reduceBy(accumulator, value);
            }
            firstOfSection = false;
        }
        // store the last accumulated value if we are splitting
        if (split) {
            put(Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level]), indexBufferKey, SEPARATOR_BYTE, lastDividingKey, SEPARATOR_BYTE, rangeEndKey]), serialize(accumulator));
            // do one final merge of the sectional accumulator into the total to determine what to return
            accumulator = await this.reduceBy(totalAccumulator, accumulator);
        }
        return { split, accumulator, version, noChildren: !split && firstOfSection };
    }
    updated(event) {
        if (event && event.sources) {
            for (let source of event.sources) {
                this.invalidateEntry(source.id, event.version);
            }
        }
        else {
            // delete the entire tree
            let level = 1;
            let hasEntries;
            do {
                const indexBufferKey = toBufferKey(this.id);
                const Class = this.constructor;
                const db = Class.db;
                const iterator = db.getRange({
                    start: Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level]), indexBufferKey, SEPARATOR_BYTE, Buffer.from([1])]),
                    end: Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([level]), indexBufferKey, SEPARATOR_BYTE, Buffer.from([255])]),
                    reverse: false,
                })[Symbol.iterator]();
                let next;
                hasEntries = false;
                while (!(next = iterator.next()).done) {
                    db.remove(next.value.key);
                    hasEntries = true;
                }
                level++;
            } while (hasEntries);
        }
        if (!this.rootLevel)
            this.rootLevel = 1;
        super.updated(event);
    }
    invalidateEntry(sourceKey, version) {
        return this.transaction(async (db, put) => {
            // get the computed entry so we know how many levels we have
            let level = this.rootLevel;
            if (!level) {
                let data = this.constructor.db.get(toBufferKey(this.id));
                if (data) {
                    const parser = createParser();
                    parser.setSource(data.slice(0, 28).toString(), 0); // the lazy version only reads the first fews bits to get the version
                    const version = parser.read();
                    level = parser.hasMoreData() && parser.read();
                }
                else {
                    return; // no entry, no levels
                }
            }
            for (let i = 1; i < level; i++) {
                let sourceKeyBuffer = toBufferKey(sourceKey);
                let [nodeToInvalidate] = await db.getRange({
                    start: Buffer.concat([REDUCED_INDEX_PREFIX_BYTE, Buffer.from([i]), toBufferKey(this.id), SEPARATOR_BYTE, sourceKeyBuffer, Buffer.from([255])]),
                    values: false,
                    reverse: true,
                    limit: 1,
                }).asArray;
                put(nodeToInvalidate.key, INVALIDATED_VALUE);
            }
            // this should happen in the super.updated call
            //put(this.id, version + ',' + level)
        });
        // rebalancing nodes will take place when we when we do the actual reduce operation
    }
    parseEntryValue(buffer) {
        if (buffer) {
            const parser = createParser();
            parser.setSource(buffer.slice(0, 28).toString(), 0); // the lazy version only reads the first fews bits to get the version
            const version = parser.read();
            this.rootLevel = parser.hasMoreData() && parser.read();
            if (parser.hasMoreData()) {
                return {
                    version,
                    data: parseLazy(buffer.slice(parser.getOffset()), parser),
                    buffer,
                };
            }
            else {
                // stored as an invalidated version
                return {
                    version,
                    buffer,
                };
            }
        }
        else {
            return {};
        }
    }
    serializeEntryValue(version, object) {
        const serializer = createSerializer();
        serializer.serialize(version);
        serializer.serialize(this.rootLevel || 1);
        if (object)
            serializer.serialize(object);
        return serializer.getSerialized();
    }
    transaction(action) {
        const Class = this.constructor;
        const db = Class.db;
        return this.currentTransaction = when(this.currentTransaction, () => {
            let operations = [];
            const put = (key, value) => {
                operations.push({
                    type: value === undefined ? 'del' : 'put',
                    key,
                    value
                });
            };
            let result = action(db, put);
            return result.then((result) => {
                return when(db.batch(operations), () => {
                    return result;
                }, (error) => {
                    console.error(error);
                    //this.currentTransaction = null
                    return result;
                });
            });
            //return result
        });
    }
}
Reduced.startingValue = undefined;
Cached.reduce = Persisted.reduce = function (name, reduceFunction) {
    let reduced = this['reduced-' + name];
    if (reduced) {
        return reduced;
    }
    reduced = this['reduced-' + name] = class extends Reduced.from(this) {
        static reduceBy(a, b) {
            return reduceFunction.call(this, a, b);
        }
    };
    Object.defineProperty(reduced, 'name', { value: this.name + '-reduced-' + name });
    return reduced;
};

const BUFFER_SIZE = 10000;
const COMMA = Buffer.from(',');
const OPEN_CURLY = Buffer.from('{');
const CLOSE_CURLY = Buffer.from('}');
const OPEN_BRACKET = Buffer.from('[');
const CLOSE_BRACKET = Buffer.from(']');
// a readable stream for serializing a set of variables to a JSON stream
class JSONStream extends stream.Readable {
    constructor(options) {
        // Calls the stream.Readable(options) constructor
        super(options);
        this.context = new RequestContext(null, options.session);
        this.context.preferJSON = true;
        this.buffer = [];
        this.bufferSize = 0;
        this.iterator = this.serialize(options.value, true);
        this.id = Math.random();
    }
    *serialize(object, containsVariables) {
        // using a generator to serialize JSON for convenience of recursive pause and resume functionality
        // serialize a value to an iterator that can be consumed by streaming API
        if (object && typeof object === 'object') {
            if (object.asJSON) {
                yield object.asJSON;
                return;
            }
            if (object[Symbol.iterator] && !object.then) {
                yield '[';
                let first = true;
                if (object[Symbol.asyncIterator] && !(object instanceof Array)) {
                    let iterator = object[Symbol.asyncIterator]();
                    let iteratorResult;
                    do {
                        iteratorResult = iterator.next();
                        if (iteratorResult.then) {
                            yield iteratorResult.then(result => {
                                iteratorResult = result;
                                return '';
                            });
                        }
                        if (iteratorResult.done) {
                            yield ']';
                            return;
                        }
                        else {
                            if (first) {
                                first = false;
                            }
                            else {
                                yield ',';
                            }
                            yield* this.serialize(iteratorResult.value);
                        }
                    } while (true);
                }
                for (let element of object) {
                    if (first) {
                        first = false;
                    }
                    else {
                        yield ',';
                    }
                    yield* this.serialize(element);
                }
                yield ']';
                return;
            }
            containsVariables = containsVariables || object.containsVariables;
            object = object.valueOf();
            if (!object)
                yield JSON.stringify(object);
            else if (object.then) {
                try {
                    yield this.context.executeWithin(() => object.then(object => this.serialize(object, containsVariables), handleError));
                }
                catch (error) {
                    yield handleError(error);
                }
            }
            else if (object.asJSON) {
                yield object.asJSON;
            }
            else if (containsVariables) {
                yield '{';
                let first = true;
                for (let key in object) {
                    if (first) {
                        first = false;
                    }
                    else {
                        yield ',';
                    }
                    yield JSON.stringify(key) + ':';
                    yield* this.serialize(object[key]);
                }
                yield '}';
            }
            else {
                yield JSON.stringify(object);
            }
        }
        else {
            yield JSON.stringify(object);
        }
    }
    _read() {
        if (this._amReading) {
            // I don't know why _read is called from within a push call, but if we are already reading, ignore the call
            return;
        }
        this._amReading = true;
        if (this.done) {
            return this.push(null);
        }
        when(this.readIterator(this.iterator), done => {
            if (done) {
                this.done = true;
                this.push(null);
            }
            else {
                this._amReading = false;
            }
        }, error => {
            console.error(error);
            this.done = true;
            this.push(error.toString());
            this.push(null);
        });
    }
    push(content) {
        if (content === null || content instanceof Buffer) {
            if (this.bufferSize > 0)
                this.flush();
            return super.push(content);
        }
        this.bufferSize += content.length || content.toString().length;
        this.buffer.push(content);
        if (this.bufferSize > BUFFER_SIZE) {
            return this.flush();
        }
        return true;
    }
    flush() {
        let pushResult = super.push(this.buffer.join(''));
        this.buffer = [];
        this.bufferSize = 0;
        return pushResult;
    }
    readIterator(iterator) {
        try { // eventually we should be able to just put this around iterator.next()
            let nextString;
            if (iterator.childIterator) {
                // resuming in a child iterator
                return when(this.readIterator(iterator.childIterator), done => {
                    if (done) {
                        iterator.childIterator = null;
                        // continue on with the current iterator
                        return this.readIterator(iterator);
                    }
                });
            }
            do {
                let stepReturn = iterator.next();
                if (stepReturn.done) {
                    return true;
                }
                nextString = stepReturn.value;
                if (nextString == null) {
                    nextString = 'null';
                }
                else {
                    if (nextString.then) {
                        this.flush();
                        return Promise.resolve(nextString).then((resolved) => {
                            if (resolved && typeof resolved.return === 'function') {
                                iterator.childIterator = resolved;
                                return this.readIterator(iterator);
                            }
                            else if (this.push(resolved + '')) {
                                return this.readIterator(iterator);
                            } // else return false
                        });
                    }
                    if (typeof nextString.return === 'function') {
                        iterator.childIterator = nextString;
                        return this.readIterator(iterator);
                    }
                }
            } while (this.push(nextString));
        }
        catch (error) {
            console.error(error);
            this.push(error.toString());
            this.push(null);
            return true;
        }
    }
}
function handleError(error) {
    console.error(error);
    return JSON.stringify(error.toString());
}
const jsonMediaType = {
    q: 0.9,
    parse: (content) => content.length > 0 ? JSON.parse(content) : undefined,
    serialize(data, connection, parameters) {
        connection.response.headers['Transfer-Encoding'] = 'chunked';
        return new JSONStream({
            value: data,
            session: connection.session
        });
    }
};

let maxRequestBody = 10000000; // default to 10MB
function bufferStream(stream) {
    return new Promise((resolve, reject) => {
        var chunks = [];
        var length = 0;
        stream.on('data', (data) => {
            chunks.push(data);
            length += data.length;
            if (length > maxRequestBody) {
                stream.connection.destroy();
                const error = new Error('Request Entity Too Large');
                error.status = 413;
                reject(error);
            }
        });
        stream.on('end', () => {
            resolve(Buffer.concat(chunks, length));
        });
    });
}

const textMediaType = {
    parse: (content) => content,
    q: 0.1,
    serialize: (content) => {
        if (content === null || content === void 0 ? void 0 : content.next)
            return stream.Readable.from(content);
        return content;
    }, // just the text
};

const htmlMediaType = {
    q: 0.2,
    serialize(data, connection, parameters) {
        // TODO: Cache this
        connection.response.headers['transfer-encoding'] = 'chunked';
        var stream = new msgpackr.PackrStream();
        stream.push(fs__default.readFileSync(require.resolve('../../client/resource-viewer.html')));
        const push = stream.push;
        stream.push = function (chunk) {
            if (chunk) {
                push.call(stream, '<script>nextData(' + JSON.stringify(chunk.toString()) + ')</script>'); // .replace(/(["/\n\r\\])/g, '\\$1'
            }
            else {
                push.call(stream, '<script>finished()</script></html>');
                push.call(stream, null);
            }
        };
        stream.end(data);
        return stream;
    }
};
function sendResourceEditor(connection) {
    connection.body = connection.response.content = fs__default.readFileSync(require.resolve('alkali/dist/index')) + '\n' +
        fs__default.readFileSync(require.resolve('dpack/dist/index')) + '\n' +
        fs__default.readFileSync(require.resolve('../../client/resource-viewer.js'));
}

const mediaTypes = new Map();
function media(connection, next) {
    let request = connection.request;
    if (connection.path.indexOf('cobase-resource-viewer') > -1) {
        return sendResourceEditor(connection);
    }
    let headers = request.headers;
    const options = {
        charset: 'utf8'
    };
    const contentType = headers['content-type'];
    if (contentType) {
        let [mimeType, optionsString] = contentType.split(/\s*;\s*/);
        if (optionsString) {
            optionsString.replace(/([^=]+)=([^;]+)/g, (t, name, value) => options[name] = value);
        }
        let parser = mediaTypes.get(mimeType);
        if (!parser || (!parser.parse && !parser.handleRequest)) {
            if (headers['content-length'] == '0') {
                parser = EMPTY_MEDIA_PARSER;
            }
            else {
                connection.status = 415;
                connection.response.content = 'Unsupported media type ' + mimeType;
                return;
            }
        }
        if (parser.handleRequest) {
            return when(parser.handleRequest(connection), () => when(next(), (returnValue) => serializer(returnValue, connection)));
        }
        return bufferStream(connection.req).then(data => {
            connection.request.data = parser.parse(data.toString(options.charset), connection);
            return when(next(), (returnValue) => serializer(returnValue, connection));
        });
    }
    return when(next(), (returnValue) => serializer(returnValue, connection));
}
function serializer(returnValue, connection) {
    var _a;
    returnValue = connection.data !== undefined ? connection.data :
        connection.response.data !== undefined ? connection.response.data : returnValue;
    if (returnValue === undefined)
        return; // nothing to serialize
    let requestHeaders = connection.request.headers;
    let acceptHeader = ((_a = connection.query) === null || _a === void 0 ? void 0 : _a.accept) || requestHeaders.accept || '*/*';
    let responseHeaders = connection.response.headers;
    responseHeaders.vary = (responseHeaders.vary ? responseHeaders.vary + ',' : '') + 'Accept';
    let bestSerializer = jsonMediaType; // default for now, TODO: return a 415
    let bestQuality = 0;
    let bestType = 'application/json'; // default
    let bestParameters;
    const acceptTypes = acceptHeader.split(/\s*,\s*/);
    for (const acceptType of acceptTypes) {
        const [type, ...parameterParts] = acceptType.split(/\s*;\s*/);
        let clientQuality = 1;
        const parameters = { q: 1 };
        for (const part of parameterParts) {
            const equalIndex = part.indexOf('=');
            parameters[part.substring(0, equalIndex)] = part.substring(equalIndex + 1);
        }
        clientQuality = +parameters.q;
        const serializer = mediaTypes.get(type);
        if (serializer) {
            const quality = (serializer.q || 1) * clientQuality;
            if (quality > bestQuality) {
                bestSerializer = serializer;
                bestType = type;
                bestQuality = quality;
                bestParameters = parameters;
            }
        }
    }
    if (connection.response.set) {
        connection.response.set('content-type', bestType.startsWith('text') ? bestType + '; charset=utf8' : bestType);
        connection.response.set('vary', 'Accept');
    }
    else {
        responseHeaders['content-type'] = bestType;
    }
    try {
        connection.response.body = connection.response.content = bestSerializer.serialize(returnValue, connection, bestParameters);
    }
    catch (error) {
        console.error(error);
        connection.response.body = connection.response.content = 'Error serializing: ' + error.toString();
    }
}
mediaTypes.set('application/json', jsonMediaType);
mediaTypes.set('text/plain', textMediaType);
mediaTypes.set('text/html', htmlMediaType);
const EMPTY_MEDIA_PARSER = {
    parse() {
    }
};

const INITIALIZING_LAST_KEY$2 = Buffer.from([1, 7]);
class Aggregator extends PersistedBase {
    static updateAggregate(previousEntry, entry) {
    }
    static forValue(id, entry) {
        return this.tryForQueueEntry(id, () => {
            this.updateAggregate(entry.previousValue, entry.value);
        });
    }
    static forQueueEntry(id) {
        return this.tryForQueueEntry(id, () => {
            return when(this.sources[0].get(id), value => this.updateAggregate(null, value));
            //				if (complete) {
            //				complete.commit()
            //		}
        });
    }
    static get(id, mode) {
        let entry = this.db.getEntry(id, mode ? 2 : 0);
        // don't use versions
        return entry && entry.value;
    }
    static fetchAllIds() {
        return [];
    }
    static from(...sources) {
        return Cached.from.apply(this, sources);
    }
    static derivedFrom(...sources) {
        for (let source of sources) {
            if (source.notifies) {
                if (!this.sources)
                    this.sources = [];
                this.sources.push(source);
            }
            else if (typeof source === 'function') {
                this.updateAggregate = source;
            }
            else {
                Object.assign(this, source);
            }
        }
        this.start();
    }
    static openDatabase() {
        this.sources[0].openChildDB(this, { cache: true });
        return false; // is not root
    }
    static getIdsFromKey(key) {
        return this.sources[0].getIdsFromKey(key);
    }
    static updateDBVersion() {
        if (!this.sources[0].wasReset) // only reindex if the source didn't do it for us
            this.db.putSync(INITIALIZING_LAST_KEY$2, this.resumeFromKey = true);
        super.updateDBVersion();
    }
    static clearEntries() {
        // don't really have any way of doing this right now
    }
    static resumeQueue() {
        this.state = 'waiting for upstream source to build';
        // explicitly wait for source to finish resuming before our own resuming
        return when(this.sources[0].resumePromise, () => super.resumeQueue());
    }
    static updated(event, by) {
        // don't do anything, we don't want these events to propagate through here, and we do indexing based on upstream queue
    }
}

function configure$1(options) {
    if (options.dbFolder || options.cacheDbFolder || options.doesInitialization !== undefined) {
        configure(options);
    }
    if (options.RequestContext) {
        setRequestContextClass(options.RequestContext);
    }
}

exports.AccessError = AccessError;
exports.Aggregator = Aggregator;
exports.Cached = Cached;
exports.ExpirationStrategy = ExpirationStrategy;
exports.Index = Index;
exports.JSONStream = JSONStream;
exports.KeyIndex = KeyIndex;
exports.Persistable = Persistable;
exports.Persisted = Persisted;
exports.Reduced = Reduced;
exports.RequestContext = RequestContext;
exports.UnauthenticatedError = UnauthenticatedError;
exports.bufferStream = bufferStream;
exports.configure = configure$1;
exports.getCurrentStatus = getCurrentStatus;
exports.media = media;
exports.mediaTypes = mediaTypes;
exports.secureAccess = secureAccess;
exports.when = when;
exports.writeCommonStructures = writeCommonStructures;
