"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ordered_binary_1 = require("ordered-binary");
const snappy_1 = require("snappy");
const WeakValueMap_1 = require("../util/WeakValueMap");
// first byte of a buffer indicates its state:
const READY_STATE = 0; // current entry and up-to-date
const INVALIDATED_STATE = 1; // has been invalidated (and a new entry has or will replace it)
const CONSIDER_FOR_COMPRESSION = 2; // current entry and up-to-date
const COMPRESSED_STATE = 3;
const HAS_COMPRESSED_STATE = 4; // not the current entry due to a compressed entry replacing it
// anything value smaller than this will be mem-copied, as it is not worth tracking the reference into mmap with a WeakValueMap
const COPY_VALUE_THRESHOLD = 100;
const COMPRESSION_THRESHOLD = 512;
class ManagedDB extends LMDB {
    constructor() {
        super(...arguments);
        this.activeBuffers = new WeakValueMap_1.default();
    }
    putWithVersion(id, value, version, metaValue, offset) {
        if (offset !== 16) {
            value = Buffer.concat([Buffer.alloc(16), value]);
        }
        value.writeUInt16BE(0, 2);
        value.writeUIntBE(version, 2, 6);
        value.writeUInt16BE(metaValue, 14); // used for level
        this.put(key, value);
    }
    invalidate(id, version, returnPreviousState) {
        return db.transaction(() => {
            let result = db.get(ordered_binary_1.toBufferKey(id));
            result.writeUIntBE(version, 8, 6); // we write directly to the shared mmap of the old entry to indicate it is invalidated (super fast)
            return result.slice(16);
        });
    }
    get(id) {
        let result = super.get(id);
        if (result.length > COPY_VALUE_THRESHOLD)
            // for larger buffers, we can save (a lot of) memory by keeping a direct reference to the shared mmap memory,
            // but, we need to track the buffer so we know if it is collected so when we advance the earliestReadTnx
            // to free old pages, we can copy if necessary if the reference is could become no longer valid.
            this.activeBuffers.set(txnId++, result);
        else
            // for smaller blocks, we mem-copy it to our own buffer, so it is our automatically handled by node/GC
            // and we don't need to track the reference to shared memory.
            result = Buffer.from(result);
        return result;
    }
    getWithVersion(id) {
        let buffer = this.get(id);
        return {
            version: buffer.readUIntBE(2, 6),
            nextVersion: buffer.readUIntBE(8, 6),
            meta: buffer.readUInt16BE(14),
            buffer: buffer.slice(16)
        };
    }
    compressUnusedEntries(id) {
        let iterator = this.iterable({})[Symbol.iterator]();
        let entry = iterator.next();
        let buffer = entry.value;
        if (buffer[0] === 0 && buffer.length > COMPRESSION_THRESHOLD) {
            buffer[0] = 1; // consider for compression
        }
        else if (state == CONSIDER_FOR_COMPRESSION) {
            snappy_1.compress(entry.value.slice(2), (compressedBuffer) => {
                this.transaction(() => {
                    let result = db.get(id);
                    if (result[0] == CONSIDER_FOR_COMPRESSION) { // make sure it hasn't changed
                        result[0] = entry.value[HAS_COMPRESSED_STATE];
                        this.put(key, Buffer.concat([Buffer.from([COMPRESSED_STATE, 0]), compressedBuffer]), {
                            noSync: true,
                        });
                    }
                });
            });
        }
    }
}
//# sourceMappingURL=managed-db.js.map